"""
The :mod:`skrmt.covariance` module implements different methods
to estimate covariance matrices.
"""

from .estimator import sample_estimator
from .estimator import fsopt_estimator
from .estimator import linear_shrinkage_estimator
from .estimator import analytical_shrinkage_estimator
from .estimator import empirical_bayesian_estimator
from .estimator import minimax_estimator

from .metrics import loss_mv, loss_frobenius
from .metrics import prial_mv


__all__ = ["sample_estimator", "fsopt_estimator",
           "linear_shrinkage_estimator",
           "analytical_shrinkage_estimator",
           "empirical_bayesian_estimator",
           "minimax_estimator",
           "loss_mv", "loss_frobenius",
           "prial_mv"]



"""
Plotting spectral laws 
=======================

In this section, we will briefly explain what is the spectral distribution
of a random matrix and how to plot it in order to study its properties.
Most ensembles spectral distribution are explained through spectral laws,
that are going to be detailed as well. 

"""

# Author: Alejandro Santorum Varela
# License: BSD 3-Clause

##############################################################################
# Spectral distribution
# ------------------------------
# 
# Most of the information of a random matrix is explained through their
# eigenvalues. Random matrix eigenvalues or random matrix spectrum follows
# a certain distribution depending on the ensemble. We would like to study
# that distribution in order to have insight of the random matrix limiting
# behaviour.
#
# We can plot the eigenvalue density of any random matrix ensemble implemented
# in scikit-rmt. However, some technical details must be taken into account.

##############################################################################
# As an example, we can try to plot the spectral density of the Gaussian
# Orthogonal Ensemble (GOE) in the interval (-10, 10).

from skrmt.ensemble import GaussianEnsemble

goe = GaussianEnsemble(beta=1, n=1000)
goe.plot_eigval_hist(bins=80, interval=(-10, 10), normalize=False)

##############################################################################
# As we can see, the histogram does not explain barely anything about
# the ensemble. That is because we are not using correctly the theory
# behind the model.
#
# Most of the ensembles need their eigenvalues to be normalized by a
# certain constant in order to perceive their limiting density properly.
# By default, the scikit-rmt is implemented to normalize eigenvalues
# with the more suitable constant depending on the ensemble.

##############################################################################
# In the former example, GOE matrices of size :math:`n \times n` should be
# normalized by :math:`1/\sqrt{n}`, which is done by default or by setting
# `normalize=True`. 

import numpy as np
from skrmt.ensemble import GaussianEnsemble

n=1000
goe = GaussianEnsemble(beta=1, n=n)
goe.plot_eigval_hist(bins=80, interval=(-10, 10), normalize=True)

##############################################################################
# Additionally, we can control the plotting interval to get a better picture
# of the spectral density. In the GOE case, the spectral distribution is
# concentrated in the (-2,2) interval.

import numpy as np
from skrmt.ensemble import GaussianEnsemble

n=1000
goe = GaussianEnsemble(beta=1, n=n)
goe.plot_eigval_hist(bins=80, interval=(-2, 2), normalize=True)

##############################################################################
# This example would be equivalent for Gaussian Unitary Ensemble (GUE) and
# Gaussian Symplectic Ensemble (GSE).

##############################################################################
# Another example would be using the Wishart Ensemble. In the following
# snippet of code Wishart Real Ensemble (WRE) is sampled.

from skrmt.ensemble import WishartEnsemble

p, n = 1000, 3000
wre = WishartEnsemble(beta=1, p=p, n=n)
wre.plot_eigval_hist(bins=80, interval=(-5,5))

##############################################################################
# By default, scikit-rmt takes charge of using the most suitable plotting
# interval and eigenvalue normalization constant.

from skrmt.ensemble import WishartEnsemble

p, n = 1000, 3000
wre = WishartEnsemble(beta=1, p=p, n=n)
wre.plot_eigval_hist(bins=80)

##############################################################################
# This example would be equivalent for Wishart Complex Ensemble (WCE) and
# Wishart Quaternion Ensemble (WQE), with the detail of controlling its
# representation interval.

##############################################################################
# So far we have analyzed spectral distributions of GOE and WRE. In the first
# example it is easy to see that its eigenvalue distribution is quite similar
# to a semicircle. Thats because Gaussian Ensemble random matrices spectrum
# follow *Wigner's Semicircle Law*. In the second example, the shown density
# has a very particular shape too, it is known as the *Marchenko-Pastur Law*.
# But, what is a spectral law?

##############################################################################
# Spectral laws
# --------------------------
# 
# Spectral laws define random matrix eigenvalue distribution when its size
# goes to infinity. So the spectral laws describes the limiting behavior
# of the spectrum of a random matrix ensemble.
#
# The most known spectral laws, and implemented by scikit-rmt, are the
# following:
#
# - **Winger Semicircle Law**: describes the limiting behaviour of Gaussian
#   Ensemble random matrix spectrum.
#
# - **Marchenko-Pastur Law**: describes the limiting behaviour of Wishart
#   Ensemble random matrix spectrum.
#
# - **Tracy-Widom Law**: describes the limiting behaviour of the largest
#   eigenvalue of a Gaussian Ensemble random matrix.

##############################################################################
# Anyone can forget about the specific plotting details and directly study
# those laws by using the functions provided in scikit-rmt.

##############################################################################
# We can easily analyze **Wigner's Semicircle Law** as follows.

from skrmt.ensemble.spectral_law import WignerSemicircleDistribution

wsd = WignerSemicircleDistribution(beta=1)
wsd.plot_empirical_pdf(n_size=2000, bins=80)

##############################################################################
# We can also study **Marchenko-Pastur Law** as by:

from skrmt.ensemble.spectral_law import MarchenkoPasturDistribution

mpd = MarchenkoPasturDistribution(ratio=1/3, beta=1)
mpd.plot_empirical_pdf(n_size=6000, bins=80)

##############################################################################
# Finally, **Tracy-Widom Law** can be represented using:

from skrmt.ensemble.spectral_law import TracyWidomDistribution

twd = TracyWidomDistribution(beta=1)
twd.plot_empirical_pdf(n_size=100, times=10000, bins=80)

##############################################################################
# Spectral laws analytical expression
# -----------------------------------
# 
# The spectral laws described so far have been proven to converge to certain
# analytical functions that defines the limiting behaviour of the eigenvalue
# distribution of the random matrices. The functions of scikit-rmt that are
# capable of plotting the spectral laws also support the representation of the
# theoretical eigenvalue pdf.

##############################################################################
# The analytical probability function for the Gaussian Ensemble, known as
# Wigner Semicircle Law, supported on :math:`[-R, R]` and centered at :math:`(0,0)`
# is :math:`f(x) = \frac{2}{\pi R^2} \sqrt{R^2 - x^2}`.

from skrmt.ensemble.spectral_law import WignerSemicircleDistribution

wsd = WignerSemicircleDistribution(beta=1)
wsd.plot_empirical_pdf(n_size=2000, bins=80, density=True, plot_law_pdf=True)

##############################################################################
# The analytical probability function for the Wishart Ensemble known as
# Marchenko-Pastur Law with parameter :math:`\lambda = p/n \in (0,1]` is
# :math:`f_{\lambda}(x) = \frac{1}{2\pi \sigma^2}\frac{\sqrt{(\lambda_+ - x)(x - \lambda_-)}}{\lambda x}`,
# where :math:`\lambda_{\pm} = \sigma^2 (1 \pm \sqrt{\lambda})^2`. 
# If :math:`\lambda > 1` then the limiting distribution has an additional
# mass probability point in the origin of size :math:`1 - \frac{1}{\lambda}`.

from skrmt.ensemble.spectral_law import MarchenkoPasturDistribution

mpd = MarchenkoPasturDistribution(ratio=1/3, beta=1)
mpd.plot_empirical_pdf(n_size=6000, bins=80, density=True, plot_law_pdf=True)

##############################################################################
# In the other hand, the Tracy-Widom Law has a complex analytical expression,
# that is the solution of a particular non-linear differential equation, described
# in detail in:
# S. Bauman. "The Tracy-Widom Distribution and its Application to Statistical Physics".
# MIT Department of Physics. 2017.
# The package scikit-rmt represents a precise
# approximation of the theoretical Tracy-Widom pdf.

from skrmt.ensemble.spectral_law import TracyWidomDistribution

twd = TracyWidomDistribution(beta=1)
twd.plot_empirical_pdf(n_size=10, times=5000, bins=80, density=True, plot_law_pdf=True)

##############################################################################
# Finally, the limiting distribution of the Manova Ensemble is not described
# by any famous Law, but its expression has been determined. This library
# provides functionality to show it on top of the empirical histogram of the
# eigenvalue spectrum.

from skrmt.ensemble.spectral_law import ManovaSpectrumDistribution

msd = ManovaSpectrumDistribution(a=3, b=3, beta=1)
msd.plot_empirical_pdf(m_size=1000, bins=80, density=True, plot_law_pdf=True)



"""Tridiagonal utils module

This module contains utilities for tridiagolization or to handle
tridiagonal matrices.

"""

import numpy as np


def tridiag_eigval_neg(tridiag_mtx):
    """Calculates number of negative eigenvalues.

    Given a tridiagonal matrix, this function calculates the
    number of negative eigenvalues using Sturm sequences.

    Args:
        tridiag_mtx (numpy array): tridiagonal matrix

    Returns:
        integer representing the number of negative eigenvalues.
    """
    n_size = len(tridiag_mtx) # number of rows = number of columns
    sturm_seq = np.zeros(n_size)

    sturm_seq[0] = tridiag_mtx[n_size-1][n_size-1] # R[0] = a_1 by default
    for i in range(1, n_size): # O(n)
        sturm_seq[i] = tridiag_mtx[n_size-i-1][n_size-i-1] - \
                        ((tridiag_mtx[n_size-i-1][n_size-i]**2)/sturm_seq[i-1])

    # number of negative values in R sequence
    return (sturm_seq<0).sum()


def tridiag_eigval_hist(tridiag_mtx, interval, bins=100, density=False):
    """Computes efficiently eigenvalue histogram.

    Calculates the eigenvalue histogram of the given matrix, using the
    specified bins between the introduced interval. The given matrix has
    to be tridiagonal, so this function builds the histogram efficiently
    using Sturm sequences, avoiding to calculate eigenvalues.

    Args:
        tridiag_mtx (numpy array): tridiagonal matrix
        interval (tuple): Delimiters (xmin, xmax) of the histogram. The lower and upper
            range of the bins. Lower and upper outliers are ignored.
        bins (int or sequence, default=100): If bins is an integer, it defines the number of
            equal-width bins in the range. If bins is a sequence, it defines the
            bin edges, including the left edge of the first bin and the right
            edge of the last bin; in this case, bins may be unequally spaced.
        density (bool, default=False): If True, draw and return a probability
            density: each bin will display the bin's raw count divided by the total
            number of counts and the bin width, so that the area under the histogram
            integrates to 1. If set to False, the absolute frequencies of the eigenvalues
            are returned.

    Returns:
        (tuple) tuple containing:
            observed (array): List of eigenvalues frequencies per bin. If density is
            True these values are the relative frequencies in order to get an area under
            the histogram equal to 1. Otherwise, this list contains the absolute
            frequencies of the eigenvalues.
            bins (array): The edges of the bins. Length nbins + 1 (nbins left edges and
            right edge of last bin)
    """
    if not isinstance(interval, tuple):
        raise ValueError("interval argument must be a tuple")

    if isinstance(bins, int):
        # calculating bin delimiters
        bin_delimiters = np.linspace(interval[0], interval[1], num=bins+1) # O(m)
    elif isinstance(bins, (tuple, list)):
        # bin delimiters directly introduced
        bin_delimiters = np.array(bins)
    else:
        raise ValueError("bins argument must be a tuple, list or integer")

    n_size = int(len(tridiag_mtx)) # number of rows = number of columns
    # + 2 because of (-inf, interval[0]) and (interval[1], inf) intervals
    histogram = np.zeros(len(bin_delimiters))

    # Building matrix A (aux_mtx) in linear time
    diag = np.arange(n_size)
    diag_1 = diag[:-1] # = np.arange(n-1)
    aux_mtx = np.zeros((n_size,n_size))
    aux_mtx[diag_1, diag_1+1] = tridiag_mtx[diag_1, diag_1+1]
    aux_mtx[diag_1+1, diag_1] = tridiag_mtx[diag_1+1, diag_1]

    prev = 0
    for i in range(bins+1): # O(m)
        aux_mtx[diag, diag] = tridiag_mtx[diag, diag] - bin_delimiters[i]
        current = tridiag_eigval_neg(aux_mtx) # O(n)
        histogram[i] = current - prev
        prev = current

    if density:
        # dont need (-inf, interval[0]) interval
        histogram[1:] = histogram[1:] / (sum(histogram[1:]) * np.diff(bin_delimiters))

    # dont want (-inf, interval[0]) interval
    return histogram[1:], bin_delimiters


def householder_reduction(mtx, ret_iterations=False):
    """Householder reduction method for tridiagonalization.

    Computes Householder reduction method for tridiagonalization. It
    transforms a given symmetric matrix in its tridiagonal form, keeping the
    same eigenvalues as the original.

    Args:
        mtx (numpy array): symmetric matrix to be tridiagonalized.
        ret_iterations (bool, default=False): If set to True, it returns
            rotation matrices used to perform the tridiagonalization.

    Returns:
        (tuple) tuple containing:
            mtx (nparray): tridiagonalized matrix.
            mtx_list (nparray): list of matrices representing the evolution of the given matrix
            after each rotation. Only returned if ret_iterations is set to True.
            rot_list (nparray): list of applied rotation matrices. Only returned if
            ret_iterations is set to True.

    References:
        - R. Hildebrand. “Householder numerically with mathematica.” 2007.
            http://buzzard.ups.edu/courses/2007spring/projects/hildebrand-paper-revised.pdf
    """
    n_size = len(mtx)

    mtx_list = [mtx]
    rot_list = []

    for j in range(n_size-2):
        if mtx[j+1, j] >= 0:
            alpha = -np.sqrt((mtx[j+1:n_size,j]**2).sum())
        else:
            alpha = np.sqrt((mtx[j+1:n_size,j]**2).sum())

        r_var = np.sqrt(alpha**2/2 - alpha/2 * mtx[j+1,j])

        x_vec = np.zeros((n_size,1))
        x_vec[j+1] = (mtx[j+1,j] - alpha)/(2*r_var)
        for k in range(j+2, n_size):
            x_vec[k] = mtx[k,j]/(2*r_var)

        rot = np.identity(n_size) - 2*np.matmul(x_vec, x_vec.transpose())
        mtx = np.matmul(rot, np.matmul(mtx, rot))

        mtx_list.append(mtx)
        rot_list.append(rot)

    if ret_iterations:
        return mtx, mtx_list, rot_list
    return mtx



"""Estimator Module

This module contains the implementation of several covariance matrix
estimators given n observations of size p.
"""

import numpy as np


def sample_estimator(data_mtx, shrink=None):
    """Estimates sample covariance matrix.

    Estimates sample covariance matrix given data matrix of size nxp.

    Args:
        data_mtx: data matrix containing n observations of size p, i.e.,
            data_mtx is a n times p matrix.
        shrink: number of degrees of freedom to substract.

    Returns:
        numpy array representing sample covariance matrix.

    References:
        Ledoit, O. and Wolf, M.
            "Analytical nonlinear shrinkage of large-dimensional covariance matrices".
            Annals of Statistics. 48.5 (2020): 3043-3065
        Numpy API documentation. numpy.cov
            https://numpy.org/doc/stable/reference/generated/numpy.cov.html
    """
    n_size = data_mtx.shape[0]

    if shrink is None:
        # demean data matrix
        data_mtx = data_mtx - data_mtx.mean(axis=0)
        # subtract one degree of freedom
        shrink=1
    # effective sample size
    n_eff=n_size-shrink
    # get sample covariance estimator
    sigma_tilde = np.matmul(data_mtx.T, data_mtx)/n_eff
    return sigma_tilde



def fsopt_estimator(data_mtx, sigma):
    """Estimates FSOpt estimator.

    Estimates finite-sample optimal estimator (FSOpt estimator), also
    written as S^*. It replaces eigenvalues of sample covariance matrix
    for a new term derived from original population covariance matrix.
    This estimator is not observable in the reality, only using Monte
    Carlo simulations.

    Args:
        data_mtx: data matrix containing n observations of size p, i.e.,
            data_mtx is a n times p matrix.
        sigma: population covariance matrix.

    Returns:
        numpy array representing sample covariance matrix.

    References:
        Ledoit, O. and Wolf, M.
            "Analytical nonlinear shrinkage of large-dimensional covariance matrices".
            Annals of Statistics. 48.5 (2020): 3043-3065
    """
    #n, p = data_mtx.shape

    sample = sample_estimator(data_mtx)

    eigvals, eigvects = np.linalg.eig(sample)
    order = np.argsort(eigvals)
    eigvals = eigvals[order]
    eigvects = eigvects[:,order]

    d_star = np.array([np.matmul(np.matmul(vec, sigma), vec.T) for vec in eigvects.T])

    # compute finite-sample optimal (FSOpt) nonlinear shrinkage estimator
    sigma_tilde = np.matmul(np.matmul(eigvects, np.diag(d_star)), eigvects.T)
    return sigma_tilde



def linear_shrinkage_estimator(data_mtx, shrink=None):
    """Estimates linear shrinkage estimator.

    When the dimension p is greater than the number of observations n,
    sample covariance matrix S is not even invertible. When the ratio p/n
    is smaller than one but not negligible, then S is invertible but it
    is not well-conditoned, i.e., inverting it amplifies estimation error.
    To get a structured well-conditioned estimator is to force the
    condition that variances are equal and covariances zero.
    Linear shrinkage estimator is a weighted average of this structured
    estimator and the sample covariance matrix.

    Args:
        data_mtx: data matrix containing n observations of size p, i.e.,
            data_mtx is a n times p matrix.
        shrink: shrinkage value. If it is not provided, this routine
            calculates the shrinkage value selected by Ledoit and Wolf
            (see reference).

    Returns:
        numpy array representing sample covariance matrix.

    References:
        Ledoit, O. and Wolf, M.
            "A well-conditioned estimator for large-dimensional covariance matrices".
            Journal of Multivariate Analysis. 88 (2004): 365-411
    """
    n_size, p_size = data_mtx.shape

    # demean data matrix
    data_mtx = data_mtx - data_mtx.mean(axis=0)
    # compute sample covariance matrix
    sample = np.matmul(data_mtx.T, data_mtx)/n_size # WATCH OUT: it is not the 'effective size'

    # compute prior
    meanvar = np.mean(np.diag(sample)) # = trace/p = sum(eigvals)/p
    prior = meanvar * np.identity(p_size)

    # use specified shrinkage value
    if shrink:
        shrinkage = shrink
    # compute shrinkage parameters
    else:
        aux_mtx = data_mtx**2
        phi_mat = np.matmul(aux_mtx.T, aux_mtx)/n_size - sample**2
        phi = np.sum(phi_mat)
        # np norm by default calculates frobenius norm for matrices and L2-norm for vects
        gamma = np.linalg.norm(sample-prior)**2
        # compute shrinkage constant
        kappa= phi/gamma
        shrinkage=max(0, min(1, kappa/n_size))

    # compute shrinkage estimator
    sigma_tilde = shrinkage*prior + (1-shrinkage)*sample
    return sigma_tilde



def analytical_shrinkage_estimator(data_mtx, shrink=None):
    """Estimates analytical shrinkage estimator.
    This estimator combines the best qualities of three different estimators:
    the speed of linear shrinkage, the accuracy of the well-known QuEST function
    and the transparency of the routine NERCOME. This estimator achieves this
    goal through nonparametric kernel estimation of the limiting spectral
    density of the sample eigenvalues and its Hilbert transform.

    Args:
        data_mtx (numpy array): data matrix containing n observations of size p, i.e.,
            data_mtx is a n times p matrix.
        shrink (integer): number of degrees of freedom to substract.

    Returns:
        numpy array representing sample covariance matrix.
    References:
        Ledoit, O. and Wolf, M.
            "Analytical nonlinear shrinkage of large-dimensional covariance matrices".
            Annals of Statistics. 48.5 (2020): 3043-3065
    """
    n_size, p_size = data_mtx.shape

    if shrink is None:
        # demean data matrix
        data_mtx = data_mtx - data_mtx.mean(axis=0)
        # subtract one degree of freedom
        shrink=1
    # effective sample size
    n_size=n_size-shrink

    # get sample eigenvalues and eigenvectors, and sort them in ascending order
    sample = np.matmul(data_mtx.T, data_mtx)/n_size
    eigvals, eigvects = np.linalg.eig(sample)
    order = np.argsort(eigvals)
    eigvals = eigvals[order]
    eigvects = eigvects[:,order]

    # compute analytical nonlinear shrinkage kernel formula
    #eigvals = eigvals[max(0,p-n):p]
    repmat_eigs = np.tile(eigvals, (min(p_size,n_size), 1)).T
    h_list = n_size**(-1/3) * repmat_eigs.T

    eigs_div = np.divide((repmat_eigs-repmat_eigs.T), h_list)

    f_tilde=(3/4/np.sqrt(5))*np.mean(np.divide(np.maximum(1-eigs_div**2/5, 0), h_list), axis=1)

    hilbert_temp = (-3/10/np.pi)*eigs_div + \
                    (3/4/np.sqrt(5)/np.pi)*(1-eigs_div**2/5)*\
                        np.log(abs((np.sqrt(5)-eigs_div)/(np.sqrt(5)+eigs_div)))
    hilbert_temp[abs(eigs_div)==np.sqrt(5)] = (-3/10/np.pi) * eigs_div[abs(eigs_div)==np.sqrt(5)]
    hilbert = np.mean(np.divide(hilbert_temp, h_list), axis=1)

    # if p <= n: (we could improve it to support p>n case)
    d_tilde = np.divide(eigvals,
                        (np.pi*(p_size/n_size)*eigvals*f_tilde)**2 + \
                            (1-(p_size/n_size)-np.pi*(p_size/n_size)*eigvals*hilbert)**2
                       )

    # compute analytical nonlinear shrinkage estimator (sigma_tilde)
    return np.matmul(np.matmul(eigvects, np.diag(d_tilde)), eigvects.T)



def empirical_bayesian_estimator(data_mtx):
    """Estimates empirical bayesian estimator.

    The empirical bayesian estimator is a linear combination of sample
    covariance matrix and the identity matrix. This estimator was
    introduced by Haff in 1980, and he suggested that this estimator
    should be used when the criterion is the mean squared error.

    Args:
        data_mtx: data matrix containing n observations of size p, i.e.,
            data_mtx is a n times p matrix.

    Returns:
        numpy array representing sample covariance matrix.

    References:
        Haff, L.R.
            "Empirical Bayes estimation of the multivariate normal covariance matrix".
            Annals of Statistics. 8 (1980): 586-597
        Ledoit, O. and Wolf, M.
            "A well-conditioned estimator for large-dimensional covariance matrices".
            Journal of Multivariate Analysis. 88 (2004): 365-411
    """
    n_size, p_size = data_mtx.shape

    sample = sample_estimator(data_mtx)

    # we consider n >= p
    det = np.linalg.det(sample)
    m_eb = det**(1/p_size)

    prior = m_eb * np.identity(p_size)
    sigma_tilde = ((p_size*n_size - 2*n_size -2)/(p_size*n_size**2))*prior + \
                    n_size/(n_size+1)*sample
    return sigma_tilde



def minimax_estimator(data_mtx):
    """Estimates minimax estimator.

    Minimax estimator has the lowest worst-case error. The minimax
    criterion is sometimes criticized as overly pessimistic, since
    it looks at the worst case only. This estimator preserves sample
    eigenvectors and replaces sample eigenvalues by (n*lambda_i)/(n+p+1-2i),
    where lambda_i (i=1,...,p) are the sample eigenvalues sorted in
    descending order.

    Args:
        data_mtx: data matrix containing n observations of size p, i.e.,
            data_mtx is a n times p matrix.

    Returns:
        numpy array representing sample covariance matrix.

    References:
        Haff, L.R.
            "Estimation of a covariance matrix under Stein’s loss".
            Annals of Statistics. 13.4 (1985): 1581-1591
        Ledoit, O. and Wolf, M.
            "A well-conditioned estimator for large-dimensional covariance matrices".
            Journal of Multivariate Analysis. 88 (2004): 365-411
    """
    n_size, p_size = data_mtx.shape

    # get sample eigenvalues and eigenvectors, and sort them in descending order
    sample = sample_estimator(data_mtx)
    eigvals, eigvects = np.linalg.eig(sample)
    order = np.argsort(eigvals) # ascending order
    order = order[::-1] # descending order (reversing it)
    eigvals = eigvals[order]
    eigvects = eigvects[:,order]

    # calculating new eigenvalues
    new_eigvals = [n_size*val/(n_size+p_size+1-2*i) for (i, val) in enumerate(eigvals)]

    # compute minimax estimator by replacing eigenvalues
    sigma_tilde = np.matmul(np.matmul(eigvects, np.diag(new_eigvals)), eigvects.T)
    return sigma_tilde



"""
Boosting density representation
===============================

Describes how speed up histogram representation of ensemble spectral density.
"""

# Author: Alejandro Santorum Varela
# License: BSD 3-Clause

import time

from skrmt.ensemble import GaussianEnsemble, WishartEnsemble

##############################################################################
# We can sample a random matrix of the Gaussian Ensemble with a relatively
# large size and plot its spectral density.

goe = GaussianEnsemble(beta=1, n=5000)

t1 = time.time()
goe.plot_eigval_hist(bins=60, interval=(-2,2))
t2 = time.time()

##############################################################################
# The time needed to compute the previous histogram is:
print(t2-t1, 'seconds')

##############################################################################
# We can speed up this procedure using the tridiagonal form of Gaussian
# Ensemble random matrices.

goe_tridiag = GaussianEnsemble(beta=1, n=5000, use_tridiagonal=True)

t1 = time.time()
goe_tridiag.plot_eigval_hist(bins=60, interval=(-2,2))
t2 = time.time()

##############################################################################
# The time needed to compute the previous histogram is:
print(t2-t1, 'seconds')

##############################################################################
# This boosting trick can be used with Wishart Ensemble too

wre = WishartEnsemble(beta=1, p=5000, n=10000)

t1 = time.time()
wre.plot_eigval_hist(bins=60, interval=(0,3))
t2 = time.time()

##############################################################################
# The time needed to compute the previous histogram is:
print(t2-t1, 'seconds')

##############################################################################
# Using tridiagonal form of the Wishart Ensemble:

wre_tridiag = WishartEnsemble(beta=1, p=5000, n=10000, use_tridiagonal=True)

t1 = time.time()
wre_tridiag.plot_eigval_hist(bins=60, interval=(0,3))
t2 = time.time()

##############################################################################
# The time needed to compute the previous histogram is:
print(t2-t1, 'seconds')




'''Gaussian Ensemble Test module

Testing GaussianEnsemble module
'''

import pytest
import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_array_equal,
)

from skrmt.ensemble import GaussianEnsemble



def test_init_exception():
    with pytest.raises(ValueError):
        _ = GaussianEnsemble(beta=3, n=100)

def test_build_tridiagonal_exception():
    with pytest.raises(ValueError):
        _ = GaussianEnsemble(beta=1, n=100, sigma=2.0, use_tridiagonal=True)


##########################################
### Gaussian Orthogonal Ensemble = GOE

def test_goe_init():
    '''Testing GOE init
    '''
    n_size = 3

    np.random.seed(1)
    goe = GaussianEnsemble(beta=1, n=n_size)

    assert goe.matrix.shape == (n_size,n_size)

    assert_almost_equal(goe.matrix, np.array([[ 2.29717124, -1.1912805 ,  0.8602944 ],
                                              [-1.1912805 ,  1.22387121, -2.16568818],
                                              [0.8602944 , -2.16568818,  0.45118942]]),
                        decimal=4)


def test_goe_symmetric():
    '''Testing that GOE matrix is symmetric
    '''
    n_size = 5
    goe = GaussianEnsemble(beta=1, n=n_size)

    m_mtx = goe.matrix
    assert (m_mtx.transpose() == m_mtx).all()



def test_goe_set_size():
    '''Testing setter to change matrix sizes of GOE
    '''
    n1_size = 3
    n2_size = 5

    goe = GaussianEnsemble(beta=1, n=n1_size)
    assert goe.n == n1_size
    assert goe.matrix.shape == (n1_size,n1_size)

    goe.set_size(n2_size, resample_mtx=False)
    assert goe.n == n2_size
    assert goe.matrix.shape == (n1_size,n1_size)

    goe.set_size(n2_size, resample_mtx=True)
    assert goe.n == n2_size
    assert goe.matrix.shape == (n2_size,n2_size)


def test_goe_build_tridiagonal():
    '''Testing tridiagonal form of GSE
    '''
    n_size = 5
    beta = 1

    np.random.seed(1)
    goe = GaussianEnsemble(beta=1, n=n_size, use_tridiagonal=True)

    np.random.seed(1)
    normals = (1/np.sqrt(2)) * np.random.normal(loc=0, scale=np.sqrt(2), size=n_size)
    dfs = np.flip(np.arange(1, n_size))
    chisqs = (1/np.sqrt(2)) * np.array([np.sqrt(np.random.chisquare(df*beta)) for df in dfs])

    for i in range(n_size):
        assert normals[i] == goe.matrix[i][i]
    for i in range(n_size-1):
        assert chisqs[i] == goe.matrix[i][i+1]
        assert chisqs[i] == goe.matrix[i+1][i]


def test_beta1_joint_eigval_pdf():
    '''Testing joint eigenvalue pdf
    '''
    n_size = 3
    goe = GaussianEnsemble(beta=1, n=n_size)

    goe.matrix = np.zeros((n_size,n_size))
    assert goe.joint_eigval_pdf() == 0.0

    goe.matrix = np.eye(n_size)
    assert goe.joint_eigval_pdf() == 0.0

    goe.matrix = 10*np.eye(n_size)
    assert goe.joint_eigval_pdf() == 0.0


def test_goe_tridiag_hist():
    '''Testing tridiagonal histogram of GOE
    '''
    n_size = 50
    goe1 = GaussianEnsemble(beta=1, n=n_size, use_tridiagonal=False)
    goe2 = GaussianEnsemble(beta=1, n=n_size, use_tridiagonal=True)

    goe1.matrix = goe2.matrix

    nbins = 10
    interval = (-2,2)
    to_norm = False # without density normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = goe1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm, normalize=False)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = goe2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm, normalize=False)
    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)

    to_norm = True # density normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = goe1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm, normalize=False)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = goe2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm, normalize=False)

    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)

    to_norm = False # no density normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = goe1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm, normalize=False)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = goe2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm, normalize=False)

    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)



##########################################
### Gaussian Unitary Ensemble = GUE

def test_gue_init():
    '''Testing GUE init
    '''
    n_size = 3

    np.random.seed(1)
    gue = GaussianEnsemble(beta=2, n=n_size)

    assert gue.matrix.shape == (n_size,n_size)

    mtx_sol = [[2.29717124+0.j, -1.1912805+1.26184983j,\
                0.8602944-0.67899889j],
               [-1.1912805-1.26184983j, 1.22387121+0.j, \
                -2.16568818+0.92362122j],
               [0.8602944+0.67899889j, -2.16568818-0.92362122j,\
                0.45118942+0.j]]

    assert_almost_equal(gue.matrix, np.array(mtx_sol),
                        decimal=4)


def test_gue_hermitian():
    '''Testing that GUE matrix is hermitian
    '''
    n_size = 5
    gue = GaussianEnsemble(beta=2, n=n_size)

    m_mtx = gue.matrix
    assert (m_mtx.transpose().conj() == m_mtx).all()


def test_gue_set_size():
    '''Testing setter to change matrix sizes of GUE
    '''
    n1_size = 5
    n2_size = 8

    gue = GaussianEnsemble(beta=2, n=n1_size)
    assert gue.n == n1_size
    assert gue.matrix.shape == (n1_size,n1_size)

    gue.set_size(n2_size, resample_mtx=False)
    assert gue.n == n2_size
    assert gue.matrix.shape == (n1_size,n1_size)

    gue.set_size(n2_size, resample_mtx=True)
    assert gue.n == n2_size
    assert gue.matrix.shape == (n2_size,n2_size)


def test_gue_build_tridiagonal():
    '''Testing tridiagonal form of GUE
    '''
    n_size = 5
    beta = 2

    np.random.seed(1)
    gue = GaussianEnsemble(beta=2, n=n_size, use_tridiagonal=True)

    np.random.seed(1)
    normals = (1/np.sqrt(2)) * np.random.normal(loc=0, scale=np.sqrt(2), size=n_size)
    dfs = np.flip(np.arange(1, n_size))
    chisqs = (1/np.sqrt(2)) * np.array([np.sqrt(np.random.chisquare(df*beta)) for df in dfs])

    for i in range(n_size):
        assert normals[i] == gue.matrix[i][i]
    for i in range(n_size-1):
        assert chisqs[i] == gue.matrix[i][i+1]
        assert chisqs[i] == gue.matrix[i+1][i]


def test_gue_tridiag_hist():
    '''Testing tridiagonal histogram of GUE
    '''
    n_size = 50
    gue1 = GaussianEnsemble(beta=2, n=n_size, use_tridiagonal=False)
    gue2 = GaussianEnsemble(beta=2, n=n_size, use_tridiagonal=True)

    gue1.matrix = gue2.matrix

    nbins = 10
    interval = (-2,2)
    to_norm = False # without density normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = gue1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm, normalize=False)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = gue2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm, normalize=False)
    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)

    to_norm = True # density normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = gue1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm, normalize=False)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = gue2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm, normalize=False)
    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)


##########################################
### Gaussian Symplectic Ensemble = GSE

def test_gse_init():
    '''Testing GSE init
    '''
    n_size = 2
    np.random.seed(1)
    gse = GaussianEnsemble(beta=4, n=n_size)

    assert gse.matrix.shape == (2*n_size,2*n_size)

    mtx_sol = [[2.29717124+0.j, -0.80605094-2.86120185j,\
                0.+0.j, -1.21019792-1.0732635j],
               [-0.80605094+2.86120185j, -1.51740678+0.j,\
                1.21019792+1.0732635j, 0.+0.j],
               [0.+0.j, 1.21019792-1.0732635j,\
                2.29717124+0.j, -0.80605094+2.86120185j],
               [-1.21019792+1.0732635j, 0.+0.j,\
                -0.80605094-2.86120185j, -1.51740678+0.j]]

    assert_almost_equal(gse.matrix, np.array(mtx_sol),
                        decimal=4)


def test_gse_hermitian():
    '''Testing that GSE matrix is hermitian
    '''
    n_size = 5
    gse = GaussianEnsemble(beta=4, n=n_size)

    m_tx = gse.matrix
    assert (m_tx.transpose().conj() == m_tx).all()


def test_gse_set_size():
    '''Testing setter to change matrix sizes of GSE
    '''
    n1_size = 4
    n2_size = 9

    gse = GaussianEnsemble(beta=4, n=n1_size)
    assert gse.n == n1_size
    assert gse.matrix.shape == (2*n1_size,2*n1_size)

    gse.set_size(n2_size, resample_mtx=False)
    assert gse.n == n2_size
    assert gse.matrix.shape == (2*n1_size,2*n1_size)

    gse.set_size(n2_size, resample_mtx=True)
    assert gse.n == n2_size
    assert gse.matrix.shape == (2*n2_size,2*n2_size)


def test_gse_build_tridiagonal():
    '''Testing tridiagonal form of GSE
    '''
    n_size = 5
    beta = 4

    np.random.seed(1)
    gse = GaussianEnsemble(beta=4, n=n_size, use_tridiagonal=True)

    np.random.seed(1)
    n_size *= 2  # WQE matrices are 2p times 2p
    normals = (1/np.sqrt(2)) * np.random.normal(loc=0, scale=np.sqrt(2), size=n_size)
    dfs = np.flip(np.arange(1, n_size))
    chisqs = (1/np.sqrt(2)) * np.array([np.sqrt(np.random.chisquare(df*beta)) for df in dfs])

    for i in range(n_size):
        assert normals[i] == gse.matrix[i][i]
    for i in range(n_size-1):
        assert chisqs[i] == gse.matrix[i][i+1]
        assert chisqs[i] == gse.matrix[i+1][i]


def test_gse_tridiag_hist():
    '''Testing tridiagonal histogram of GSE
    '''
    n_size = 50
    gse1 = GaussianEnsemble(beta=4, n=n_size, use_tridiagonal=False)
    gse2 = GaussianEnsemble(beta=4, n=n_size, use_tridiagonal=True)

    gse1.matrix = gse2.matrix

    nbins = 10
    interval = (-2,2)
    to_norm = False # without density normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = gse1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm, normalize=False)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = gse2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm, normalize=False)
    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)

    to_norm = True # density normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = gse1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm, normalize=False)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = gse2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm, normalize=False)
    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)



"""Gaussian Ensemble Module

This module contains the implementation of the Gaussian Ensemble, also
known as Hermite Ensemble. This ensemble of random matrices contains
mainly three sub-ensembles: Gaussian Orthogonal Ensemble (GOE),
Gaussian Unitary Ensemble (GUE) and Gaussian Symplectic Ensemble (GSE).

"""

import numpy as np
from scipy import sparse, special
import matplotlib.pyplot as plt

from ._base_ensemble import _Ensemble
from .tridiagonal_utils import tridiag_eigval_hist

#########################################################################
### Gaussian Ensemble = Hermite Ensemble

class GaussianEnsemble(_Ensemble):
    """General Gaussian Ensemble class.

    This class contains common attributes and methods for all the
    gaussian ensembles. Gaussian Ensembles are divided in:
    - Gaussian Orthogonal Ensemble (GOE, beta=1): the distribution of the
    matrices of this ensemble are invariant under orthogonal conjugation,
    i.e., if X is in GOE(n) and O is an orthogonal matrix, then
    O*X*O^T is equally distributed as X.
    - Gaussian Unitary Ensemble (GUE, beta=2): the distribution of
    the matrices of this ensemble are invariant under unitary conjugation,
    i.e., if X is in GUE(n) and O is an unitary matrix, then O*X*O^T
    is equally distributed as X.
    - Gaussian Symplectic Ensemble (GSE, beta=4): the distribution of
    the matrices of this ensemble are invariant under conjugation
    by the symplectic group.

    Attributes:
        matrix (numpy array): instance of the GOE, GUE or GSE random
            matrix ensemble of size n times n if it is GOE or GUE, or
            of size 2n times 2n if it is GSE.
        beta (int): descriptive integer of the gaussian ensemble type.
            For GOE beta=1, for GUE beta=2, for GSE beta=4.
        n (int): random matrix size. Gaussian ensemble matrices are
            squared matrices. GOE and GUE are of size n times n,
            and GSE are of size 2n times 2n.
        use_tridiagonal (bool): if set to True, Gaussian Ensemble
            matrices are sampled in its tridiagonal form, which has the same
            eigenvalues than its standard form. Otherwise, it is sampled using
            its standard form.
        sigma (float): scale (standard deviation) of the random entries of the
            sampled matrix.

    References:
        - Albrecht, J. and Chan, C.P. and Edelman, A.
            "Sturm sequences and random eigenvalue distributions".
            Foundations of Computational Mathematics. 9.4 (2008): 461-483.
        - Dumitriu, I. and Edelman, A. "Matrix Models for Beta Ensembles".
            Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

    """

    def __init__(self, beta, n, use_tridiagonal=False, sigma=1.0):
        """Constructor for GaussianEnsemble class.

        Initializes an instance of this class with the given parameters.

        Args:
            beta (int, default=1): descriptive integer of the gaussian ensemble type.
                For GOE beta=1, for GUE beta=2, for GSE beta=4.
            n (int): random matrix size. Gaussian ensemble matrices are
                squared matrices. GOE and GUE are of size n times n,
                and GSE are of size 2n times 2n.
            use_tridiagonal (bool, default=False): if set to True, Gaussian Ensemble
                matrices are sampled in its tridiagonal form, which has the same
                eigenvalues than its standard form. Otherwise, it is sampled using
                its standard form.
            sigma (float, 1.0): scale (standard deviation) of the random entries of the
                sampled matrix.

        """
        if beta not in [1,2,4]:
            raise ValueError(f"Invalid beta: {beta}. Beta value has to be 1, 2 or 4.")

        super().__init__()
        # pylint: disable=invalid-name
        self.n = n
        self.beta = beta
        self.use_tridiagonal = use_tridiagonal
        self.sigma = sigma
        self._eigvals = None
        self.matrix = self.sample()

        # default eigenvalue normalization constant
        if self.use_tridiagonal:
            self.eigval_norm_const = 1/np.sqrt(self.n) if self.beta==4 else 1/np.sqrt(self.n/2)
        else:
            self.eigval_norm_const = 1/np.sqrt(self.n)


    def set_size(self, n, resample_mtx=True):
        # pylint: disable=arguments-differ
        """Setter of matrix size.

        Sets the matrix size. Useful if it has been initialized with a different value.

        Args:
            n (int): new random matrix size. Gaussian ensemble matrices are
                squared matrices. GOE and GUE are of size n times n, and
                GSE are of size 2n times 2n.
            resample_mtx (bool, default=True): If set to True, the ensemble matrix is
                resampled with the new dimensions.

        """
        self.n = n
        if resample_mtx:
            self.matrix = self.sample()

    # pylint: disable=inconsistent-return-statements
    def sample(self):
        """Samples new Gaussian Ensemble random matrix.

        The sampling algorithm depends on the specification of
        use_tridiagonal parameter. If use_tridiagonal is set to True,
        a Gaussian Ensemble random matrix in its tridiagonal form
        is sampled. Otherwise, it is sampled using the standard
        form.

        Returns:
            numpy array containing new matrix sampled.

        References:
            - Dumitriu, I. and Edelman, A. "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.
        """
        if self.use_tridiagonal:
            return self.sample_tridiagonal()

        if self.beta == 1:
            return self._sample_goe()
        if self.beta == 2:
            return self._sample_gue()
        if self.beta == 4:
            return self._sample_gse()

    def _sample_goe(self):
        # n by n matrix of random Gaussians
        mtx = np.random.randn(self.n,self.n) * self.sigma
        # symmetrize matrix
        self.matrix = (mtx + mtx.transpose())/np.sqrt(2)
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def _sample_gue(self):
        size = self.n
        # n by n random complex matrix
        mtx = np.random.randn(size,size)*self.sigma + 1j*np.random.randn(size,size)*self.sigma
        # hermitian matrix
        self.matrix = (mtx + mtx.transpose().conj())/np.sqrt(2)
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def _sample_gse(self):
        size = self.n
        # n by n random complex matrix
        x_mtx = np.random.randn(size,size)*self.sigma + 1j*np.random.randn(size,size)*self.sigma
        # another n by n random complex matrix
        y_mtx = np.random.randn(size,size)*self.sigma + 1j*np.random.randn(size,size)*self.sigma
        # [X Y; -conj(Y) conj(X)]
        mtx = np.block([
                       [x_mtx               , y_mtx],
                       [-np.conjugate(y_mtx), np.conjugate(x_mtx)]
                        ])
        # hermitian matrix
        self.matrix = (mtx + mtx.transpose().conj())/np.sqrt(2)
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def sample_tridiagonal(self):
        '''Samples a Gaussian Ensemble random matrix in its tridiagonal form.

        Samples a random matrix of the specified Gaussian Ensemble (remember,
        beta=1 is GOE, beta=2 is GUE and beta=4 is GSE) in its tridiagonal
        form.

        Returns:
            numpy array containing new matrix sampled.

        References:
            - Albrecht, J. and Chan, C.P. and Edelman, A.
                "Sturm sequences and random eigenvalue distributions".
                Foundations of Computational Mathematics. 9.4 (2008): 461-483.
            - Dumitriu, I. and Edelman, A. "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        '''
        if self.sigma != 1.0:
            raise ValueError("Error: cannot sample tridiagonal random matrix using non-unitary scale"
                             f" (sigma = {self.sigma}).\n"
                             "\t Set `sigma=1.0` (default) or deactivate tridiagonal sampling.")

        size = 2*self.n if self.beta==4 else self.n
        # sampling diagonal normals
        normals = (1/np.sqrt(2)) * np.random.normal(loc=0, scale=np.sqrt(2), size=size)
        # sampling chi-squares
        dfs = np.flip(np.arange(1, size))
        chisqs = (1/np.sqrt(2)) * \
                 np.array([np.sqrt(np.random.chisquare(df*self.beta)) for df in dfs])
        # inserting diagonals
        diagonals = [chisqs, normals, chisqs]
        mtx = sparse.diags(diagonals, [-1, 0, 1])
        # converting to numpy array
        self.matrix = mtx.toarray()
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def eigvals(self, normalize=False):
        """Computes the random matrix eigenvalues.

        Calculates the random matrix eigenvalues using numpy standard procedure.
        If the matrix ensemble is symmetric, a faster algorithm is used.

        Returns:
            numpy array with the calculated eigenvalues.

        """
        norm_const = self.eigval_norm_const if normalize else 1.0

        if self._eigvals is not None:
            return norm_const * self._eigvals

        # always storing non-normalized eigenvalues
        self._eigvals = np.linalg.eigvalsh(self.matrix)
        return norm_const * self._eigvals

    def eigval_hist(self, bins, interval=None, density=False, normalize=True, avoid_img=False):
        if self.use_tridiagonal:
            if normalize:
                return tridiag_eigval_hist(
                    self.eigval_norm_const * self.matrix,
                    bins=bins,
                    interval=interval,
                    density=density,
                )
            return tridiag_eigval_hist(self.matrix, bins=bins, interval=interval, density=density)

        return super().eigval_hist(bins, interval=interval, density=density,
                                   normalize=normalize, avoid_img=avoid_img)

    def plot_eigval_hist(self, bins=100, interval=None, density=False, normalize=True, fig_path=None):
        """Computes and plots the histogram of the matrix eigenvalues.

        Calculates and plots the histogram of the current sampled matrix eigenvalues.
        Gaussian (Hermite) ensemble and Wishart (Laguerre) ensemble have improved
        routines to avoid calculating the eigenvalues, so the histogram
        is built using certain techniques to boost efficiency.

        Args:
            bins (int or sequence, default=100): If bins is an integer, it defines the number of
                equal-width bins in the range. If bins is a sequence, it defines the
                bin edges, including the left edge of the first bin and the right
                edge of the last bin; in this case, bins may be unequally spaced.
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram.
                The lower and upper range of the bins. Lower and upper outliers are ignored.
            density (bool, default=False): If True, draw and return a probability
                density: each bin will display the bin's raw count divided by the total
                number of counts and the bin width, so that the area under the histogram
                integrates to 1. If set to False, the absolute frequencies of the eigenvalues
                are returned.
            normalize (bool, default=True): Whether to normalize the computed eigenvalues
                by the default normalization constant (see references). Defaults to True, i.e.,
                the eigenvalues are normalized. Normalization makes the eigenvalues to be in the
                same support independently of the sample size.
            fig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.

        References:
            - Albrecht, J. and Chan, C.P. and Edelman, A.
                "Sturm sequences and random eigenvalue distributions".
                Foundations of Computational Mathematics. 9.4 (2008): 461-483.
            - Dumitriu, I. and Edelman, A.
                "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        """
        if interval is None:
            wsl_radius = 2*np.sqrt(self.beta)*self.sigma
            interval = (-wsl_radius, wsl_radius)

        if self.use_tridiagonal:
            # pylint: disable=too-many-arguments
            observed, bins = self.eigval_hist(
                bins=bins, interval=interval, density=density, normalize=normalize
            )

            width = bins[1]-bins[0]
            plt.bar(bins[:-1], observed, width=width, align='edge')
            plt.title("Eigenvalue histogram", fontweight="bold")
            plt.xlabel("x")
            plt.ylabel("density")
            # Saving plot or showing it
            if fig_path:
                plt.savefig(fig_path, dpi=1000)
            else:
                plt.show()

        else:
            # pylint: disable=too-many-arguments
            super().plot_eigval_hist(
                bins=bins,
                interval=interval,
                density=density,
                normalize=normalize,
                fig_path=fig_path,
            )

    def joint_eigval_pdf(self, eigvals=None):
        '''Computes joint eigenvalue pdf.

        Calculates joint eigenvalue probability density function given an array of
        eigenvalues. If the array of eigenvalues is not provided, the current random
        matrix sample (so its eigenvalues) is used. This function depends on beta,
        i.e., in the sub-Gaussian ensemble.

        Args:
            eigvals (np.ndarray, default=None): numpy array with the values (eigenvalues)
                to evaluate the joint pdf in.

        Returns:
            real number. Value of the joint pdf of the eigenvalues.

        References:
            - Dumitriu, I. and Edelman, A.
                "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        '''
        if eigvals is None:
            # calculating eigenvalues
            eigvals = self.eigvals()
        n_eigvals = len(eigvals)

        # calculating Hermite eigval pdf constant depeding on beta
        const_beta = (2*np.pi)**(-self.n/2)
        for j in range(self.n):
            const_beta *= special.gamma(1 + self.beta/2)/special.gamma(1 + self.beta*j/2)
        # calculating prod
        pdf = 1
        for j in range(n_eigvals):
            for i in range(j):
                pdf *= np.abs(eigvals[i] - eigvals[j])**self.beta
        # calculating exponential term
        exp_val = np.exp(-np.sum((eigvals**2)/2))
        # calculating Hermite eigval pdf
        return const_beta * pdf * exp_val



'''Manova Ensemble Test module

Testing ManovaEnsemble module
'''

import numpy as np
from numpy.testing import (
    assert_almost_equal,
)

from skrmt.ensemble import ManovaEnsemble


##########################################
### Manova Real Ensemble = MRE

def test_manova_real_init():
    '''Testing MRE init
    '''
    m_size = 3
    n1_size = 5
    n2_size = 10

    np.random.seed(1)
    mre = ManovaEnsemble(beta=1, m=m_size, n1=n1_size, n2=n2_size)

    assert mre.matrix.shape == (m_size,m_size)

    assert_almost_equal(mre.matrix, np.array([[0.30694308, -0.20376032, 0.12255971],
                                              [-0.15432555, 0.57689371, -0.1121342 ],
                                              [0.20435358, -0.32120057, 0.26201473]]),
                        decimal=7)


def test_mre_set_size():
    '''Testing setter to change matrix sizes of MRE
    '''
    m1_size, n11_size, n12_size = 3, 5, 8
    m2_size, n21_size, n22_size = 2, 4, 6

    mre = ManovaEnsemble(beta=1, m=m1_size, n1=n11_size, n2=n12_size)
    assert mre.m == m1_size
    assert mre.n1 == n11_size
    assert mre.n2 == n12_size
    assert mre.matrix.shape == (m1_size,m1_size)

    mre.set_size(m=m2_size, n1=n21_size, n2=n22_size, resample_mtx=False)
    assert mre.m == m2_size
    assert mre.n1 == n21_size
    assert mre.n2 == n22_size
    assert mre.matrix.shape == (m1_size,m1_size)

    mre.set_size(m=m2_size, n1=n21_size, n2=n22_size, resample_mtx=True)
    assert mre.m == m2_size
    assert mre.n1 == n21_size
    assert mre.n2 == n22_size
    assert mre.matrix.shape == (m2_size,m2_size)


def test_beta1_joint_eigval_pdf():
    '''Testing joint eigenvalue pdf
    '''
    m_size, n1_size, n2_size = 4, 6, 8
    mre = ManovaEnsemble(beta=1, m=m_size, n1=n1_size, n2=n2_size)

    mre.matrix = np.zeros((m_size,m_size))
    assert mre.joint_eigval_pdf() == 0.0

    mre.matrix = np.eye(m_size)
    assert mre.joint_eigval_pdf() == 0.0


##########################################
### Manova Complex Ensemble = MCE

def test_manova_complex_init():
    '''Testing MCE init
    '''
    m_size = 2
    n1_size = 1
    n2_size = 10

    np.random.seed(1)
    mce = ManovaEnsemble(beta=2, m=m_size, n1=n1_size, n2=n2_size)

    assert mce.matrix.shape == (m_size,m_size)

    assert_almost_equal(mce.matrix, np.array([[0.1027935+0.00612663j, -0.00875725+0.08423236j],
                                              [-0.01070604-0.07368927j, 0.06093032-0.00612663j]]),
                        decimal=7)


def test_mce_set_size():
    '''Testing setter to change matrix sizes of MCE
    '''
    m1_size, n11_size, n12_size = 4, 6, 8
    m2_size, n21_size, n22_size = 3, 9, 12

    mce = ManovaEnsemble(beta=2, m=m1_size, n1=n11_size, n2=n12_size)
    assert mce.m == m1_size
    assert mce.n1 == n11_size
    assert mce.n2 == n12_size
    assert mce.matrix.shape == (m1_size,m1_size)

    mce.set_size(m=m2_size, n1=n21_size, n2=n22_size, resample_mtx=False)
    assert mce.m == m2_size
    assert mce.n1 == n21_size
    assert mce.n2 == n22_size
    assert mce.matrix.shape == (m1_size,m1_size)

    mce.set_size(m=m2_size, n1=n21_size, n2=n22_size, resample_mtx=True)
    assert mce.m == m2_size
    assert mce.n1 == n21_size
    assert mce.n2 == n22_size
    assert mce.matrix.shape == (m2_size,m2_size)


##########################################
### Manova Quaternion Ensemble = MQE

def test_manova_quatern_init():
    '''Testing MQE init
    '''
    m_size = 2
    n1_size = 10
    n2_size = 20

    np.random.seed(1)
    mqe = ManovaEnsemble(beta=4, m=m_size, n1=n1_size, n2=n2_size)

    assert mqe.matrix.shape == (2*m_size,2*m_size)

    mtx_sol = [[0.36695654+0.01615832j, 0.05808572-0.08831798j, \
                0.00268224-0.00172003j, -0.01413889-0.0228631j],
               [0.05675906+0.06962803j, 0.30129103-0.01333821j, \
                0.0081629+0.01974845j, -0.00731881-0.00630646j],
               [-0.00268224-0.00172003j, 0.01413889-0.0228631j, \
                0.36695654-0.01615832j, 0.05808572+0.08831798j],
               [-0.0081629+0.01974845j, 0.00731881-0.00630646j, \
                0.05675906-0.06962803j, 0.30129103+0.01333821j]]

    assert_almost_equal(mqe.matrix, np.array(mtx_sol), decimal=7)


def test_mqe_set_size():
    '''Testing setter to change matrix sizes of MQE
    '''
    m1_size, n11_size, n12_size = 2, 5, 7
    m2_size, n21_size, n22_size = 4, 5, 6

    mqe = ManovaEnsemble(beta=4, m=m1_size, n1=n11_size, n2=n12_size)
    assert mqe.m == m1_size
    assert mqe.n1 == n11_size
    assert mqe.n2 == n12_size
    assert mqe.matrix.shape == (2*m1_size,2*m1_size)

    mqe.set_size(m=m2_size, n1=n21_size, n2=n22_size, resample_mtx=False)
    assert mqe.m == m2_size
    assert mqe.n1 == n21_size
    assert mqe.n2 == n22_size
    assert mqe.matrix.shape == (2*m1_size,2*m1_size)

    mqe.set_size(m=m2_size, n1=n21_size, n2=n22_size, resample_mtx=True)
    assert mqe.m == m2_size
    assert mqe.n1 == n21_size
    assert mqe.n2 == n22_size
    assert mqe.matrix.shape == (2*m2_size,2*m2_size)



"""Circular Ensemble Module

This module contains the implementation of the Circular Ensemble.
This ensemble of random matrices contains mainly three sub-ensembles:
Circular Orthogonal Ensemble (COE), Circular Unitary Ensemble (CUE)
and Circular Symplectic Ensemble (CSE).

"""

import numpy as np
import matplotlib.pyplot as plt
from scipy import special

from ._base_ensemble import _Ensemble



def _sample_haar_mtx(size):
    """Samples Haar-distributed matrices.

    Samples Haar-distributed matrices that are useful to generate
    random matrices for COE, CUE and CSE ensembles.

    Args:
        n (int): matrix size.

    Returns:
        numpy array containing Haar-distributed random matrix.
    """
    # n by n random complex matrix
    x_mtx = np.random.randn(size,size) + (0+1j)*np.random.randn(size,size)
    # orthonormalizing matrix using QR algorithm
    q_mtx, _ = np.linalg.qr(x_mtx)
    # the resulting Q is Haar-distributed
    return q_mtx


#########################################################################
### Circular Ensemble

class CircularEnsemble(_Ensemble):
    """General Circular Ensemble class.

    This class contains common attributes and methods for all the
    Circular ensembles. Circular Ensembles are divided in:
    - Circular Orthogonal Ensemble (COE, beta=1): the distribution
    of the matrices of this ensemble are invariant under orthogonal
    conjugation, i.e., if X is in COE(n) and O is an orthogonal matrix,
    then O*X*O^T is equally distributed as X.
    - Circular Unitary Ensemble (CUE, beta=2): the distribution of
    the matrices of this ensemble are invariant under unitary
    conjugation, i.e., if X is in CUE(n) and O is an unitary matrix,
    then O*X*O^T is equally distributed as X.
    - Circular Symplectic Ensemble (CSE, beta=4): the distribution
    of the matrices of this ensemble are invariant under conjugation
    by the symplectic group.

    Attributes:
        matrix (numpy array): instance of the COE, CUE or CSE random
            matrix ensemble of size n times n if it is COE or CUE, or
            of size 2n times 2n if it is CSE.
        beta (int): descriptive integer of the gaussian ensemble type.
            For COE beta=1, for CUE beta=2, for CSE beta=4.
        n (int): random matrix size. Circular ensemble matrices are
            squared matrices. COE and CUE are of size n times n,
            and CSE are of size 2n times 2n.

    References:
        - Killip, R. and Zozhan, R.
            Matrix Models and Eigenvalue Statistics for Truncations of
            Classical Ensembles of Random Unitary Matrices.
            Communications in Mathematical Physics. 349 (2017): 991-1027.
        - "Circular ensemble". Wikipedia.
            en.wikipedia.org/wiki/Circular_ensemble

    """

    def __init__(self, beta, n):
        """Constructor for CircularEnsemble class.

        Initializes an instance of this class with the given parameters.

        Args:
            beta (int): descriptive integer of the Circular ensemble type.
                For COE beta=1, for CUE beta=2, for CSE beta=4.
            n (int): random matrix size. Circular ensemble matrices are
                squared matrices. COE and CUE are of size n times n,
                and CSE are of size 2n times 2n.

        """
        super().__init__()
        # pylint: disable=invalid-name
        self.n = n
        self.beta = beta
        self._eigvals = None
        self.matrix = self.sample()

    def set_size(self, n, resample_mtx=False):
        # pylint: disable=arguments-differ
        """Setter of matrix size.

        Sets the matrix size. Useful if it has been initialized with a different value.

        Args:
            n (int): random matrix size. Circular ensemble matrices are
                squared matrices. COE and CUE are of size n times n,
                and CSE are of size 2n times 2n.

        """
        self.n = n
        if resample_mtx:
            self.matrix = self.sample()

    # pylint: disable=inconsistent-return-statements
    def sample(self):
        """Samples new Circular Ensemble random matrix.

        The sampling algorithm depends on the specification of
        beta parameter. If beta=1, COE matrix is sampled; if
        beta=2 CUE matrix is sampled and if beta=4
        CSE matrix is sampled.

        Returns:
            numpy array containing new matrix sampled.

        References:
            - Killip, R. and Zozhan, R.
                Matrix Models and Eigenvalue Statistics for Truncations of
                Classical Ensembles of Random Unitary Matrices.
                Communications in Mathematical Physics. 349 (2017): 991-1027.
            - "Circular ensemble". Wikipedia.
                en.wikipedia.org/wiki/Circular_ensemble

        """
        if self.beta == 1:
            return self._sample_coe()
        if self.beta == 2:
            return self._sample_cue()
        if self.beta == 4:
            return self._sample_cse()

    def _sample_coe(self):
        # sampling unitary Haar-distributed matrix
        u_mtx = _sample_haar_mtx(self.n)
        # mapping to Circular Orthogonal Ensemble
        self.matrix = np.matmul(u_mtx.transpose(), u_mtx)
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def _sample_cue(self):
        # sampling unitary Haar-distributed matrix
        self.matrix = _sample_haar_mtx(self.n)
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def _sample_cse(self):
        # sampling unitary Haar-distributed matrix of size 2n
        u_mtx = _sample_haar_mtx(2*self.n)
        # mapping to Circular Symplectic Ensemble
        j_mtx = self._build_j_mtx()
        # U_R = J * U^T * J^T
        u_r_aux = np.matmul(j_mtx, u_mtx.transpose())
        u_r_mtx = np.matmul(u_r_aux, j_mtx.transpose())
        # A = U^R * U
        self.matrix = np.matmul(u_r_mtx, u_mtx)
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def _build_j_mtx(self):
        """Creates an useful matrix to sample CSE matrices.

        Creates matrix J of zeros but with the upper-diagonal
        set to -1 and the lower-diagonal set to 1. This matrix
        is useful in the sampling algorithm of CSE matrices.

        Args:
            n (int): matrix size.

        Returns:
            numpy array containing J matrix.

        References:
            - Killip, R. and Zozhan, R.
                Matrix Models and Eigenvalue Statistics for Truncations of
                Classical Ensembles of Random Unitary Matrices.
                Communications in Mathematical Physics. 349 (2017): 991-1027.
            - "Circular ensemble". Wikipedia.
                en.wikipedia.org/wiki/Circular_ensemble
        """
        size = 2*self.n
        j_mtx = np.zeros((size,size))
        # selecting indices
        inds = np.arange(size-1)
        # selecting upper-diagonal indices
        j_mtx[inds, inds+1] = -1
        # selecting lower-diagonal indices
        j_mtx[inds+1, inds] = 1
        return j_mtx

    def eigvals(self, normalize=False):
        """Calculates the random matrix eigenvalues.

        Calculates the random matrix eigenvalues using numpy standard procedure.
        If the matrix ensemble is symmetric, a faster algorithm is used.

        Returns:
            numpy array with the calculated eigenvalues.

        """
        norm_const = self.eigval_norm_const if normalize else 1.0
        if self._eigvals is not None:
            return norm_const * self._eigvals

        if self.beta == 1:
            # using eigvalsh because it's known all eigenvalues are real
            self._eigvals = np.linalg.eigvalsh(self.matrix)
        else:
            # using eigvals since some eigenvalues could be imaginary
            self._eigvals = np.linalg.eigvals(self.matrix)

        return norm_const * self._eigvals

    def plot_eigval_hist(self, bins, interval=None, density=False, normalize=True, fig_path=None):
        """Computes and plots the histogram of the matrix eigenvalues.

        Calculates and plots the histogram of the current sampled matrix eigenvalues.
        It is important to underline that this function works with real and complex
        eigenvalues: if the matrix eigenvalues are complex, they are plotted in the
        complex plane next to a heap map to study eigenvalue density.

        Args:
            bins (int or sequence): If bins is an integer, it defines the number of
                equal-width bins in the range. If bins is a sequence, it defines the
                bin edges, including the left edge of the first bin and the right
                edge of the last bin; in this case, bins may be unequally spaced.
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram.
                The lower and upper range of the bins. Lower and upper outliers are ignored.
            density (bool, default=False): If True, draw and return a probability
                density: each bin will display the bin's raw count divided by the total
                number of counts and the bin width, so that the area under the histogram
                integrates to 1. If set to False, the absolute frequencies of the eigenvalues
                are returned.
            normalize (bool, default=True): Whether to normalize the computed eigenvalues
                by the default normalization constant (see references). Defaults to True, i.e.,
                the eigenvalues are normalized. Normalization makes the eigenvalues to be in the
                same support independently of the sample size.
            fig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.

        References:
            - Killip, R. and Zozhan, R.
                Matrix Models and Eigenvalue Statistics for Truncations of
                Classical Ensembles of Random Unitary Matrices.
                Communications in Mathematical Physics. 349 (2017): 991-1027.

        """
        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-locals
        if self.beta == 1:
            return super().plot_eigval_hist(
                bins=bins, interval=interval, density=density, normalize=normalize, fig_path=fig_path
            )

        if (interval is not None) and not isinstance(interval, tuple):
            raise ValueError("interval argument must be a tuple (or None)")

        eigvals = self.eigvals()
        xvals = eigvals.real
        yvals = eigvals.imag

        if interval is None:
            rang_val = self.beta/2
            rang = ((-rang_val, rang_val), (-rang_val, rang_val))
            extent = [-rang_val, rang_val, -rang_val, rang_val]
        else:
            rang = (interval, interval)
            extent = [interval[0], interval[1], interval[0], interval[1]]

        fig, axes = plt.subplots(nrows=1, ncols=2)
        fig.set_figheight(5)
        fig.set_figwidth(13)
        fig.subplots_adjust(hspace=.5)

        axes[0].set_xlim(rang[0][0], rang[0][1])
        axes[0].set_ylim(rang[1][0], rang[1][1])
        axes[0].plot(xvals, yvals, 'o')
        axes[0].set_title('Complex plane')
        axes[0].set_xlabel('real')
        axes[0].set_ylabel('imaginary')

        h2d,_,_,img = axes[1].hist2d(xvals, yvals, range=rang,
                                   cmap=plt.cm.get_cmap('nipy_spectral'))
        fig.colorbar(img, ax=axes[1])
        axes[1].cla()
        axes[1].imshow(h2d.transpose(), origin='lower', interpolation="bilinear", extent=extent)
        axes[1].set_title('Eigenvalue heatmap')
        axes[1].set_xlabel('real')
        axes[1].set_ylabel('imaginary')

        plt.suptitle("Complex eigenvalues histogram", fontweight="bold")

        # Saving plot or showing it
        if fig_path:
            plt.savefig(fig_path, dpi=600)
        else:
            plt.show()


    def joint_eigval_pdf(self, eigvals=None):
        '''Computes joint eigenvalue pdf.

        Calculates joint eigenvalue probability density function given an array of
        eigenvalues. If the array of eigenvalues is not provided, the current random
        matrix sample (so its eigenvalues) is used. This function depends on beta,
        i.e., in the sub-Circular ensemble.

        Args:
            eigvals (np.ndarray, default=None): numpy array with the values (eigenvalues)
                to evaluate the joint pdf in.

        Returns:
            real number. Value of the joint pdf of the eigenvalues.

        References:
            - Killip, R. and Zozhan, R.
                Matrix Models and Eigenvalue Statistics for Truncations of
                Classical Ensembles of Random Unitary Matrices.
                Communications in Mathematical Physics. 349 (2017): 991-1027.
            - "Circular ensemble". Wikipedia.
                en.wikipedia.org/wiki/Circular_ensemble

        '''
        if eigvals is None:
            # calculating eigenvalues
            eigvals = self.eigvals()
        n_eigvals = len(eigvals)

        # calculating Circular eigval pdf constant depeding on beta
        const_beta = (2*np.pi)**self.n * \
                     special.gamma(1 + self.n*self.beta/2)/(special.gamma(1 + self.beta/2)**self.n)

        # calculating prod
        pdf = 1
        for k in range(n_eigvals):
            for i in range(k):
                complex_num = complex(0, np.exp(eigvals[i])) - complex(0,np.exp(eigvals[k]))
                pdf *= np.abs(complex_num)**self.beta
        # calculating Circular eigval pdf
        return (1/const_beta) * pdf



"""
The :mod:`skrmt.ensemble` module includes models based on
random matrix ensembles.
"""

from .gaussian_ensemble import GaussianEnsemble
from .wishart_ensemble import WishartEnsemble
from .manova_ensemble import ManovaEnsemble
from .circular_ensemble import CircularEnsemble

from .spectral_law import WignerSemicircleDistribution
from .spectral_law import MarchenkoPasturDistribution
from .spectral_law import TracyWidomDistribution
from .spectral_law import ManovaSpectrumDistribution

from .tridiagonal_utils import tridiag_eigval_neg
from .tridiagonal_utils import tridiag_eigval_hist
from .tridiagonal_utils import householder_reduction


__all__ = ["GaussianEnsemble", "WishartEnsemble",
           "ManovaEnsemble", "CircularEnsemble",
           "WignerSemicircleDistribution",
           "MarchenkoPasturDistribution",
           "TracyWidomDistribution",
           "ManovaSpectrumDistribution",
           "tridiag_eigval_neg",
           "tridiag_eigval_hist",
           "householder_reduction"
        ]



'''Circular Ensemble Test module

Testing CircularEnsemble module
'''

import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_array_equal,
)

from skrmt.ensemble import CircularEnsemble


##########################################
### Circular Orthogonal Ensemble = COE

def test_coe_init():
    '''Testing COE init
    '''
    n_size = 3

    np.random.seed(1)
    coe = CircularEnsemble(beta=1, n=n_size)

    assert coe.matrix.shape == (n_size,n_size)

    mtx_sol = [[0.66482467-0.48190422j, -0.4274577+0.09781321j, -0.05592612-0.36105573j],
               [-0.4274577+0.09781321j, -0.80810768+0.1559807j, 0.36012685+0.02555646j],
               [-0.05592612-0.36105573j, 0.36012685+0.02555646j, 0.8408391-0.17075173j]]

    assert_almost_equal(coe.matrix, np.array(mtx_sol), decimal=7)


def test_coe_symmetric():
    '''Testing that COE matrix is symmetric
    '''
    n_size = 5
    coe = CircularEnsemble(beta=1, n=n_size)

    mtx = coe.matrix
    assert (mtx.transpose() == mtx).all()


def test_coe_set_size():
    '''Testing setter to change matrix sizes of COE
    '''
    n1_size = 3
    n2_size = 5

    coe = CircularEnsemble(beta=1, n=n1_size)
    assert coe.n == n1_size
    assert coe.matrix.shape == (n1_size,n1_size)

    coe.set_size(n2_size, resample_mtx=False)
    assert coe.n == n2_size
    assert coe.matrix.shape == (n1_size,n1_size)

    coe.set_size(n2_size, resample_mtx=True)
    assert coe.n == n2_size
    assert coe.matrix.shape == (n2_size,n2_size)


def test_coe_eigvals():
    '''Testing all eigenvalues of a COE matrix are real
    '''
    n_size = 5
    coe = CircularEnsemble(beta=1, n=n_size)

    vals = coe.eigvals()

    assert_array_equal(vals.imag, 0.0)


def test_beta1_joint_eigval_pdf():
    '''Testing joint eigenvalue pdf
    '''
    n_size = 3
    coe = CircularEnsemble(beta=1, n=n_size)

    coe.matrix = np.zeros((n_size,n_size))
    assert coe.joint_eigval_pdf() == 0.0

    coe.matrix = np.eye(n_size)
    assert coe.joint_eigval_pdf() == 0.0

    coe.matrix = 10*np.eye(n_size)
    assert coe.joint_eigval_pdf() == 0.0


##########################################
### Circular Unitary Ensemble = CUE

def test_cue_init():
    '''Testing CUE init
    '''
    n_size = 3

    np.random.seed(1)
    cue = CircularEnsemble(beta=2, n=n_size)

    assert cue.matrix.shape == (n_size,n_size)

    mtx_sol = [[-0.56689951+0.08703072j, 0.00821467-0.66547876j, -0.38691909+0.28002633j],
               [0.37446802+0.1125242j, -0.23983754+0.18549333j, -0.86852536-0.02908408j],
               [-0.60894251+0.38386407j, 0.1958485 +0.65328713j, -0.12797213+0.01788349j]]

    assert_almost_equal(cue.matrix, np.array(mtx_sol), decimal=7)


def test_cue_set_size():
    '''Testing setter to change matrix sizes of CUE
    '''
    n1_size = 5
    n2_size = 10

    cue = CircularEnsemble(beta=2, n=n1_size)
    assert cue.n == n1_size
    assert cue.matrix.shape == (n1_size,n1_size)

    cue.set_size(n2_size, resample_mtx=False)
    assert cue.n == n2_size
    assert cue.matrix.shape == (n1_size,n1_size)

    cue.set_size(n2_size, resample_mtx=True)
    assert cue.n == n2_size
    assert cue.matrix.shape == (n2_size,n2_size)


def test_cue_eigvals():
    '''Testing all eigenvalues of a CUE matrix have module 1
    '''
    n_size = 5
    cue = CircularEnsemble(beta=2, n=n_size)

    vals = cue.eigvals()

    mods = np.absolute(vals)
    assert_almost_equal(mods, 1.0, decimal=12)


def test_beta2_joint_eigval_pdf():
    '''Testing joint eigenvalue pdf
    '''
    n_size = 3
    cue = CircularEnsemble(beta=2, n=n_size)

    cue.matrix = np.zeros((n_size,n_size))
    assert cue.joint_eigval_pdf() == 0.0

    cue.matrix = np.eye(n_size)
    assert cue.joint_eigval_pdf() == 0.0

    cue.matrix = 10*np.eye(n_size)
    assert cue.joint_eigval_pdf() == 0.0


##########################################
### Circular Symplectic Ensemble = CSE

def test_cse_init():
    '''Testing CSE init
    '''
    n_size = 2
    np.random.seed(1)
    cse = CircularEnsemble(beta=4, n=n_size)

    assert cse.matrix.shape == (2*n_size,2*n_size)

    mtx_sol = [[4.43078888e-01-8.29771592e-01j, -6.00429653e-17+1.43650461e-17j,\
                9.66935997e-02+7.40605433e-01j, 3.43966100e-01-3.01198149e-02j],
               [4.56225923e-01+3.23872465e-02j, 3.46385288e-01-1.57037703e+00j, \
                4.56225923e-01+3.23872465e-02j, -1.22164739e+00-7.84199458e-01j],
               [-9.56297276e-01+7.17094554e-01j, -3.43966100e-01+3.01198149e-02j, \
                6.11735400e-01-6.90830132e-02j, -3.43966100e-01+3.01198149e-02j],
               [-4.56225923e-01-3.23872465e-02j, 9.66935997e-02+7.40605433e-01j, \
                1.94983679e-17+3.89651615e-18j, 7.08429000e-01+6.71522420e-01j]]

    assert_almost_equal(cse.matrix, np.array(mtx_sol), decimal=7)


def test_cse_set_size():
    '''Testing setter to change matrix sizes of CSE
    '''
    n1_size = 4
    n2_size = 9

    cse = CircularEnsemble(beta=4, n=n1_size)
    assert cse.n == n1_size
    assert cse.matrix.shape == (2*n1_size,2*n1_size)

    cse.set_size(n2_size, resample_mtx=False)
    assert cse.n == n2_size
    assert cse.matrix.shape == (2*n1_size,2*n1_size)

    cse.set_size(n2_size, resample_mtx=True)
    assert cse.n == n2_size
    assert cse.matrix.shape == (2*n2_size,2*n2_size)


def test_beta4_joint_eigval_pdf():
    '''Testing joint eigenvalue pdf
    '''
    n_size = 3
    cse = CircularEnsemble(beta=4, n=n_size)

    cse.matrix = np.zeros((n_size,n_size))
    assert cse.joint_eigval_pdf() == 0.0

    cse.matrix = np.eye(n_size)
    assert cse.joint_eigval_pdf() == 0.0

    cse.matrix = 10*np.eye(n_size)
    assert cse.joint_eigval_pdf() == 0.0



'''Spectral Law Test module

Testing law module for spectral law simulations
'''

import os
import pytest
import shutil
import numpy as np

from skrmt.ensemble import WignerSemicircleDistribution
from skrmt.ensemble import MarchenkoPasturDistribution
from skrmt.ensemble import TracyWidomDistribution
from skrmt.ensemble import ManovaSpectrumDistribution
from skrmt.ensemble.spectral_law import _indicator


TMP_DIR_PATH = os.path.join(os.getcwd(), "skrmt/ensemble/tests/tmp")


@pytest.fixture(scope="module", autouse=True)
def _setup_tmp_dir(request):
    '''Function that is run before all tests in this script.

    It creates a temporary folder in order to store useful files
    for the following tests.
    '''
    # if the directory already exists, it is deleted
    if os.path.exists(TMP_DIR_PATH):
        shutil.rmtree(TMP_DIR_PATH)
    # creating temporary directory  
    os.mkdir(TMP_DIR_PATH)

    # specifying a function that will be run after all tests are completed
    request.addfinalizer(_remove_tmp_dir)


def _remove_tmp_dir():
    '''Function that removes the created temporary directory.

    The function is run when all tests in this module are completed.
    '''
    shutil.rmtree(TMP_DIR_PATH)



class TestWignerSemicircleDistribution:

    def test_wsd_init_success(self):
        '''Testing WignerSemicircleDistribution init
        '''
        beta = 4
        sigma = 2

        wsd = WignerSemicircleDistribution(beta=beta, sigma=sigma)

        assert wsd.beta == beta
        assert wsd.center == 0.0
        assert wsd.sigma == sigma
        assert wsd.radius == 2.0 * np.sqrt(beta) * sigma
    
    def test_wsd_init_raise(self):
        '''Testing WignerSemicircleDistribution init raising exception
        '''
        with pytest.raises(ValueError):
            _ = WignerSemicircleDistribution(beta=3)

    def test_wsd_rvs_success(self):
        '''Testing WignerSemicircleDistribution random variates (sampling)
        '''
        beta = 1
        size = 5
        wsd1 = WignerSemicircleDistribution(beta=beta)
        samples = wsd1.rvs(size=size)
        assert len(samples == size)

        beta = 4
        size = 5
        wsd4 = WignerSemicircleDistribution(beta=beta)
        samples = wsd4.rvs(size=size)
        assert len(samples == size)

        size = 10
        samples = wsd4.rvs(size=size)
        assert len(samples == size)
    
    def test_wsd_rvs_raise(self):
        '''Testing WignerSemicircleDistribution random variates (sampling)
        raising an exception because of invalid argument
        '''
        with pytest.raises(ValueError):
            wsd = WignerSemicircleDistribution(beta=1)
            wsd.rvs(-5)
    
    def test_wsd_pdf(self):
        '''Testing WignerSemicircleDistribution pdf
        '''
        beta = 4
        sigma = 2
        center = 0
        wsd = WignerSemicircleDistribution(beta=beta, sigma=sigma, center=center)

        assert wsd.pdf(center) > 0.0
        assert wsd.pdf(center + wsd.radius + 0.1) == 0.0
        assert wsd.pdf(center - wsd.radius - 0.01) == 0.0

        center = 10
        wsd = WignerSemicircleDistribution(beta=beta, sigma=sigma, center=center)

        assert wsd.pdf(center) > 0.0
        assert wsd.pdf(center + wsd.radius + 0.1) == 0.0
        assert wsd.pdf(center - wsd.radius - 0.01) == 0.0
    
    def test_wsd_cdf(self):
        '''Testing WignerSemicircleDistribution cdf
        '''
        beta = 4
        sigma = 2
        center = 0
        wsd = WignerSemicircleDistribution(beta=beta, sigma=sigma, center=center)

        assert wsd.cdf(center) > 0.0
        assert wsd.cdf(center + wsd.radius + 0.1) == 1.0
        assert wsd.cdf(center - wsd.radius - 0.01) == 0.0

        center = 10
        wsd = WignerSemicircleDistribution(beta=beta, sigma=sigma, center=center)

        assert wsd.cdf(center) > 0.0
        assert wsd.cdf(center + wsd.radius + 0.1) == 1.0
        assert wsd.cdf(center - wsd.radius - 0.01) == 0.0
    
    def test_wsd_plot_pdf(self):
        '''Testing WignerSemicircleDistribution plot pdf
        '''
        fig_name = "test_wsd_pdf_wo_interval.png"
        wsd = WignerSemicircleDistribution()
        wsd.plot_pdf(savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

        fig_name = "test_wsd_pdf_w_interval.png"
        wsd = WignerSemicircleDistribution()
        wsd.plot_pdf(interval=(-2,2), savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_cdf(self):
        '''Testing WignerSemicircleDistribution plot cdf
        '''
        fig_name = "test_wsd_cdf_wo_interval.png"
        wsd = WignerSemicircleDistribution()
        wsd.plot_cdf(savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

        fig_name = "test_wsd_cdf_w_interval.png"
        wsd = WignerSemicircleDistribution()
        wsd.plot_cdf(interval=(-2,2), savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_pdf_raise(self):
        '''Testing WignerSemicircleDistribution pdf raising exception
        '''
        with pytest.raises(ValueError):
            wsd = WignerSemicircleDistribution(beta=1)
            wsd.plot_pdf(interval=1)

    def test_wsd_plot_goe_abs_freq(self):
        fig_name = "test_wsl_goe_absfreq.png"
        wsd = WignerSemicircleDistribution(beta=1)
        wsd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_gue_abs_freq(self):
        fig_name = "test_wsl_gue_absfreq.png"
        wsd = WignerSemicircleDistribution(beta=2)
        wsd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_gse_abs_freq(self):
        fig_name = "test_wsl_gse_absfreq.png"
        wsd = WignerSemicircleDistribution(beta=4)
        wsd.plot_empirical_pdf(
            n_size=50,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_goe_normalized(self):
        fig_name = "test_wsl_goe_norm.png"
        wsd = WignerSemicircleDistribution(beta=1)
        wsd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_gue_normalized(self):
        fig_name = "test_wsl_gue_norm.png"
        wsd = WignerSemicircleDistribution(beta=2)
        wsd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_gse_normalized(self):
        fig_name = "test_wsl_gse_norm.png"
        wsd = WignerSemicircleDistribution(beta=4)
        wsd.plot_empirical_pdf(
            n_size=50,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_goe_theoretical(self):
        fig_name = "test_wsl_goe_theory.png"
        wsd = WignerSemicircleDistribution(beta=1)
        wsd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_gue_theoretical(self):
        fig_name = "test_wsl_gue_theory.png"
        wsd = WignerSemicircleDistribution(beta=2)
        wsd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_gse_theoretical(self):
        fig_name = "test_wsl_gse_theory.png"
        wsd = WignerSemicircleDistribution(beta=4)
        wsd.plot_empirical_pdf(
            n_size=50,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_wsd_plot_size_exception(self):
        with pytest.raises(ValueError):
            wsd = WignerSemicircleDistribution(beta=1)
            wsd.plot_empirical_pdf(n_size=0)
    
    def test_wsd_plot_ensemble_exception(self):
        with pytest.raises(ValueError):
            wsd = WignerSemicircleDistribution(beta=0)
            wsd.plot_empirical_pdf()
    


class TestMarchenkoPasturDistribution:

    def test_mpd_init_success(self):
        '''Testing MarchenkoPasturDistribution init
        '''
        beta = 4
        ratio = 1/2
        sigma = 2.0

        mpd = MarchenkoPasturDistribution(beta=beta, ratio=ratio, sigma=sigma)

        assert mpd.beta == beta
        assert mpd.ratio == ratio
        assert mpd.sigma == sigma
        assert mpd.lambda_minus == beta * sigma**2 * (1 - np.sqrt(ratio))**2
        assert mpd.lambda_plus == beta * sigma**2 * (1 + np.sqrt(ratio))**2
        assert mpd._var == beta * sigma**2
        assert mpd._wishart_ens is None
    
    def test_mpd_init_raise(self):
        '''Testing MarchenkoPasturDistribution init raising exception
        '''
        with pytest.raises(ValueError):
            _ = MarchenkoPasturDistribution(ratio=1, beta=3)
        
        with pytest.raises(ValueError):
            _ = MarchenkoPasturDistribution(ratio=0)

    def test_mpd_rvs_success(self):
        '''Testing MarchenkoPasturDistribution random variates (sampling)
        '''
        beta = 1
        ratio = 1/3
        size = 5
        mpd1 = MarchenkoPasturDistribution(beta=beta, ratio=ratio)
        samples = mpd1.rvs(size=size)
        assert len(samples == size)

        beta = 4
        ratio = 1/3
        size = 5
        mpd4 = MarchenkoPasturDistribution(beta=beta, ratio=ratio)
        samples = mpd4.rvs(size=size)
        assert len(samples == size)

        size = 10
        samples = mpd4.rvs(size=size)
        assert len(samples == size)
    
    def test_mpd_rvs_raise(self):
        '''Testing MarchenkoPasturDistribution random variates (sampling) raising
        an exception because of an invalid argument
        '''
        with pytest.raises(ValueError):
            mpd = MarchenkoPasturDistribution(beta=1, ratio=1)
            mpd.rvs(-5)
    
    def test_mpd_pdf(self):
        '''Testing MarchenkoPasturDistribution pdf
        '''
        beta = 4
        ratio = 1/3
        sigma = 1
        mpd = MarchenkoPasturDistribution(beta=beta, ratio=ratio, sigma=sigma)

        middle = np.mean([mpd.lambda_minus, mpd.lambda_plus])
        assert mpd.pdf(middle) > 0.0
        assert mpd.pdf(mpd.lambda_plus + 0.1) == 0.0
        assert mpd.pdf(mpd.lambda_minus - 0.01) == 0.0

        sigma = 10
        mpd = MarchenkoPasturDistribution(beta=beta, ratio=ratio, sigma=sigma)

        middle = np.mean([mpd.lambda_minus, mpd.lambda_plus])
        assert mpd.pdf(middle) > 0.0
        assert mpd.pdf(mpd.lambda_plus + 0.1) == 0.0
        assert mpd.pdf(mpd.lambda_minus - 0.01) == 0.0
    
    def test_mpd_cdf(self):
        '''Testing MarchenkoPasturDistribution cdf
        '''
        beta = 1
        ratio = 1/3
        sigma = 1
        mpd = MarchenkoPasturDistribution(beta=beta, ratio=ratio, sigma=sigma)

        middle = np.mean([mpd.lambda_minus, mpd.lambda_plus])
        assert mpd.cdf(middle) > 0.0
        assert mpd.cdf(mpd.lambda_plus + 0.1) == 1.0
        assert mpd.cdf(mpd.lambda_minus - 0.01) == 0.0

        sigma = 2
        ratio = 2
        mpd = MarchenkoPasturDistribution(beta=beta, ratio=ratio, sigma=sigma)

        middle = np.mean([mpd.lambda_minus, mpd.lambda_plus])
        assert mpd.cdf(middle) > 0.0
        assert mpd.cdf(mpd.lambda_plus + 0.1) == 1.0
        assert mpd.cdf(mpd.lambda_minus - 0.1) == 0.0
    
    def test_mpd_plot_pdf(self):
        '''Testing MarchenkoPasturDistribution plot pdf
        '''
        fig_name = "test_mpd_pdf_wo_interval.png"
        mpd = MarchenkoPasturDistribution(ratio=1/3)
        mpd.plot_pdf(savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

        fig_name = "test_mpd_pdf_w_interval.png"
        mpd = MarchenkoPasturDistribution(ratio=1/3)
        mpd.plot_pdf(interval=(-1,10), savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_cdf(self):
        '''Testing MarchenkoPasturDistribution plot cdf
        '''
        fig_name = "test_mpd_cdf_wo_interval.png"
        mpd = MarchenkoPasturDistribution(ratio=1/3)
        mpd.plot_cdf(savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

        fig_name = "test_mpd_cdf_w_interval.png"
        mpd = MarchenkoPasturDistribution(ratio=1/3)
        mpd.plot_cdf(interval=(-1,10), savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

    def test_mpd_plot_wre_abs_freq(self):
        fig_name = "test_mpl_wre_absfreq.png"
        mpd = MarchenkoPasturDistribution(beta=1, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

    def test_mpd_plot_wce_abs_freq(self):
        fig_name = "test_mpl_wce_absfreq.png"
        mpd = MarchenkoPasturDistribution(beta=2, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wqe_abs_freq(self):
        fig_name = "test_mpl_wqe_absfreq.png"
        mpd = MarchenkoPasturDistribution(beta=4, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wre_normalized(self):
        fig_name = "test_mpl_wre_norm.png"
        mpd = MarchenkoPasturDistribution(beta=1, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

    def test_mpd_plot_wce_normalized(self):
        fig_name = "test_mpl_wce_norm.png"
        mpd = MarchenkoPasturDistribution(beta=2, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wqe_normalized(self):
        fig_name = "test_mpl_wqe_norm.png"
        mpd = MarchenkoPasturDistribution(beta=4, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wre_theoretical(self):
        fig_name = "test_mpl_wre_theory.png"
        mpd = MarchenkoPasturDistribution(beta=1, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

    def test_mpd_plot_wce_theoretical(self):
        fig_name = "test_mpl_wce_theory.png"
        mpd = MarchenkoPasturDistribution(beta=2, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wqe_theoretical(self):
        fig_name = "test_mpl_wqe_theory.png"
        mpd = MarchenkoPasturDistribution(beta=4, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wre_ratio_ge1(self):
        fig_name = "test_mpl_wre_ratio_ge1.png"
        mpd = MarchenkoPasturDistribution(beta=1, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wre_theoretical_ratio_ge1(self):
        fig_name = "test_mpl_wre_theory_ratio_ge1.png"
        mpd = MarchenkoPasturDistribution(beta=1, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wre_interval(self):
        fig_name = "test_mpl_wre_interval.png"
        mpd = MarchenkoPasturDistribution(beta=1, ratio=1/3)
        mpd.plot_empirical_pdf(
            n_size=100,
            bins=100,
            interval=(0,10),
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wre_emppdf_p_size(self):
        fig_name = "test_mpd_wre_emppdf_p_size.png"
        mpd = MarchenkoPasturDistribution(beta=1, ratio=1/3)
        mpd.plot_empirical_pdf(
            p_size=50,
            n_size=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wce_emppdf_p_size(self):
        fig_name = "test_mpd_wce_emppdf_p_size.png"
        mpd = MarchenkoPasturDistribution(beta=2, ratio=1/3)
        mpd.plot_empirical_pdf(
            p_size=50,
            n_size=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wqe_emppdf_p_size(self):
        fig_name = "test_mpd_wqe_emppdf_p_size.png"
        mpd = MarchenkoPasturDistribution(beta=4, ratio=1/3)
        mpd.plot_empirical_pdf(
            p_size=50,
            n_size=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wre_emppdf_ratio_g_1(self):
        fig_name = "test_mpd_plot_wre_emppdf_ratio_g_1.png"
        mpd = MarchenkoPasturDistribution(beta=1, ratio=2)
        mpd.plot_empirical_pdf(
            n_size=50,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_wre_emppdf_ratio_g_1_density(self):
        fig_name = "test_mpd_plot_wre_emppdf_ratio_g_1_density.png"
        mpd = MarchenkoPasturDistribution(beta=1, ratio=2)
        mpd.plot_empirical_pdf(
            n_size=50,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_mpd_plot_mpl_size_exception(self):
        with pytest.raises(ValueError):
            mpd = MarchenkoPasturDistribution(beta=1, ratio=1/3)
            mpd.plot_empirical_pdf(n_size=0)
    
    def test_mpd_plot_mpl_ensemble_exception(self):
        with pytest.raises(ValueError):
            mpd = MarchenkoPasturDistribution(beta=0, ratio=1/3)
            mpd.plot_empirical_pdf()



class TestTracyWidomDistribution:

    def test_twd_init_success(self):
        '''Testing TracyWidomDistribution init
        '''
        beta = 4

        twd = TracyWidomDistribution(beta=beta)

        assert twd.beta == beta
        assert twd.tw_approx is not None
    
    def test_twd_init_raise(self):
        '''Testing TracyWidomDistribution init raising exception
        '''
        with pytest.raises(ValueError):
            _ = TracyWidomDistribution(beta=3)

    def test_twd_rvs_success(self):
        '''Testing TracyWidomDistribution random variates (sampling)
        '''
        beta = 1
        size = 5
        twd1 = TracyWidomDistribution(beta=beta)
        samples = twd1.rvs(size=size)
        assert len(samples == size)

        beta = 4
        size = 5
        twd4 = TracyWidomDistribution(beta=beta)
        samples = twd4.rvs(size=size)
        assert len(samples == size)

        size = 10
        samples = twd4.rvs(size=size)
        assert len(samples == size)
    
    def test_twd_rvs_raise(self):
        '''Testing TracyWidomDistribution random variates (sampling) raising
        an exception because of an invalid argument
        '''
        with pytest.raises(ValueError):
            twd = TracyWidomDistribution(beta=1)
            twd.rvs(-5)

        with pytest.raises(ValueError):
            twd = TracyWidomDistribution(beta=1)
            twd.rvs(size=5, mtx_size=0)
    
    def test_twd_pdf(self):
        '''Testing TracyWidomDistribution pdf
        '''
        beta = 4
        twd = TracyWidomDistribution(beta=beta)
        assert twd.pdf(-1) > 0.0
        assert twd.pdf(100) < 1e-10
        assert twd.pdf(-100) < 1e-10
    
    def test_twd_cdf(self):
        '''Testing TracyWidomDistribution cdf
        '''
        beta = 4
        twd = TracyWidomDistribution(beta=beta)
        assert twd.cdf(-1) > 0.0
        assert twd.cdf(100) > 0.99999
        assert twd.pdf(-100) < 1e-10
    
    def test_twd_plot_pdf(self):
        '''Testing TracyWidomDistribution plot pdf
        '''
        fig_name = "test_twd_pdf_wo_interval.png"
        twd = TracyWidomDistribution()
        twd.plot_pdf(savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

        fig_name = "test_twd_pdf_w_interval.png"
        twd = TracyWidomDistribution()
        twd.plot_pdf(interval=(-5,5), savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_cdf(self):
        '''Testing TracyWidomDistribution plot cdf
        '''
        fig_name = "test_twd_cdf_wo_interval.png"
        twd = TracyWidomDistribution()
        twd.plot_cdf(savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

        fig_name = "test_twd_cdf_w_interval.png"
        twd = TracyWidomDistribution()
        twd.plot_cdf(interval=(-5,5), savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_goe_abs_freq(self):
        fig_name = "test_twl_goe_abs_freq.png"
        twd = TracyWidomDistribution(beta=1)
        twd.plot_empirical_pdf(
            n_size=50,
            times=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_gue_abs_freq(self):
        fig_name = "test_twl_gue_abs_freq.png"
        twd = TracyWidomDistribution(beta=2)
        twd.plot_empirical_pdf(
            n_size=50,
            times=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_gse_abs_freq(self):
        fig_name = "test_twl_gse_abs_freq.png"
        twd = TracyWidomDistribution(beta=4)
        twd.plot_empirical_pdf(
            n_size=25,
            times=100,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_goe_normalized(self):
        fig_name = "test_twl_goe_normalized.png"
        twd = TracyWidomDistribution(beta=1)
        twd.plot_empirical_pdf(
            n_size=50,
            times=100,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_gue_normalized(self):
        fig_name = "test_twl_gue_normalized.png"
        twd = TracyWidomDistribution(beta=2)
        twd.plot_empirical_pdf(
            n_size=50,
            times=100,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_gse_normalized(self):
        fig_name = "test_twl_gse_normalized.png"
        twd = TracyWidomDistribution(beta=4)
        twd.plot_empirical_pdf(
            n_size=25,
            times=100,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_goe_theoretical(self):
        fig_name = "test_twl_goe_theory.png"
        twd = TracyWidomDistribution(beta=1)
        twd.plot_empirical_pdf(
            n_size=50,
            times=100,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_gue_theoretical(self):
        fig_name = "test_twl_gue_theory.png"
        twd = TracyWidomDistribution(beta=2)
        twd.plot_empirical_pdf(
            n_size=50,
            times=100,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_gse_theoretical(self):
        fig_name = "test_twl_gse_theory.png"
        twd = TracyWidomDistribution(beta=4)
        twd.plot_empirical_pdf(
            n_size=25,
            times=100,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_twd_plot_size_exception(self):
        with pytest.raises(ValueError):
            twd = TracyWidomDistribution(beta=1)
            twd.plot_empirical_pdf(n_size=0)
    
    def test_twd_plot_ensemble_exception(self):
        with pytest.raises(ValueError):
            twd = TracyWidomDistribution(beta=0)
            twd.plot_empirical_pdf(n_size=10)



class TestManovaSpectrumDistribution:

    def test_msd_init_success(self):
        '''Testing ManovaSpectrumDistribution init
        '''
        beta = 4
        a = 3
        b = 3

        msd = ManovaSpectrumDistribution(beta=beta, a=a, b=b)

        assert msd.beta == beta
        assert msd.a == a
        assert msd.b == b

        assert msd.lambda_term1 == np.sqrt((a/(a+b)) * (1 - (1/(a+b))))
        assert msd.lambda_term2 == np.sqrt((1/(a+b)) * (1 - (a/(a+b))))
        assert msd.lambda_minus == (msd.lambda_term1 - msd.lambda_term2)**2
        assert msd.lambda_plus == (msd.lambda_term1 + msd.lambda_term2)**2
        assert msd._manova_ens is None
    
    def test_msd_init_raise(self):
        '''Testing ManovaSpectrumDistribution init raising exception
        '''
        with pytest.raises(ValueError):
            _ = ManovaSpectrumDistribution(a=1, b=1, beta=3)
        
        with pytest.raises(ValueError):
            _ = ManovaSpectrumDistribution(a=0, b=0, beta=1)

    def test_msd_rvs_success(self):
        '''Testing ManovaSpectrumDistribution random variates (sampling)
        '''
        beta = 1
        a = b = 3
        size = 5
        msd1 = ManovaSpectrumDistribution(beta=beta, a=a, b=b)
        samples = msd1.rvs(size=size)
        assert len(samples == size)

        beta = 4
        a = b = 3
        size = 5
        msd4 = ManovaSpectrumDistribution(beta=beta, a=a, b=b)
        samples = msd4.rvs(size=size)
        assert len(samples == size)

        size = 10
        samples = msd4.rvs(size=size)
        assert len(samples == size)
    
    def test_msd_rvs_raise(self):
        '''Testing ManovaSpectrumDistribution random variates (sampling) raising an
        exception because of an invalid argument
        '''
        with pytest.raises(ValueError):
            msd = ManovaSpectrumDistribution(beta=1, a=1, b=1)
            msd.rvs(-5)
    
    def test_msd_pdf(self):
        '''Testing ManovaSpectrumDistribution pdf
        '''
        beta = 4
        a = b = 2
        msd = ManovaSpectrumDistribution(beta=beta, a=a, b=b)

        middle = np.mean([msd.lambda_minus, msd.lambda_plus])
        assert msd.pdf(middle) > 0.0
        assert msd.pdf(msd.lambda_plus + 0.1) == 0.0
        assert msd.pdf(msd.lambda_minus - 0.01) == 0.0
    
    def test_msd_cdf(self):
        '''Testing ManovaSpectrumDistribution cdf
        '''
        beta = 4
        a = b = 2
        msd = ManovaSpectrumDistribution(beta=beta, a=a, b=b)

        middle = np.mean([msd.lambda_minus, msd.lambda_plus])
        assert msd.cdf(middle) > 0.0
        assert msd.cdf(msd.lambda_plus + 0.1) == 1.0
        assert msd.cdf(msd.lambda_minus - 0.01) == 0.0
    
    def test_msd_plot_pdf(self):
        '''Testing ManovaSpectrumDistribution plot pdf
        '''
        fig_name = "test_msd_pdf_wo_interval.png"
        msd = ManovaSpectrumDistribution(a=3, b=3)
        msd.plot_pdf(savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

        fig_name = "test_msd_pdf_w_interval.png"
        msd = ManovaSpectrumDistribution(a=3, b=3)
        msd.plot_pdf(interval=(-1,2), savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_cdf(self):
        '''Testing ManovaSpectrumDistribution plot cdf
        '''
        fig_name = "test_msd_cdf_wo_interval.png"
        msd = ManovaSpectrumDistribution(a=3, b=3)
        msd.plot_cdf(savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

        fig_name = "test_msd_cdf_w_interval.png"
        msd = ManovaSpectrumDistribution(a=3, b=3)
        msd.plot_cdf(interval=(-1,2), savefig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mre_abs_freq(self):
        fig_name = "test_msd_mre_absfreq.png"
        msd = ManovaSpectrumDistribution(beta=1, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=40,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

    def test_msd_plot_mce_abs_freq(self):
        fig_name = "test_msd_mce_absfreq.png"
        msd = ManovaSpectrumDistribution(beta=2, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=40,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mqe_abs_freq(self):
        fig_name = "test_msd_mqe_absfreq.png"
        msd = ManovaSpectrumDistribution(beta=4, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=40,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mre_normalized(self):
        fig_name = "test_msd_mre_norm.png"
        msd = ManovaSpectrumDistribution(beta=1, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=40,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

    def test_msd_plot_mce_normalized(self):
        fig_name = "test_msd_mce_norm.png"
        msd = ManovaSpectrumDistribution(beta=2, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=40,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mqe_normalized(self):
        fig_name = "test_msd_mqe_norm.png"
        msd = ManovaSpectrumDistribution(beta=4, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=40,
            bins=100,
            density=True,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mre_theoretical(self):
        fig_name = "test_msd_mre_theory.png"
        msd = ManovaSpectrumDistribution(beta=1, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=40,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

    def test_msd_plot_mce_theoretical(self):
        fig_name = "test_msd_mce_theory.png"
        msd = ManovaSpectrumDistribution(beta=2, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=40,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mqe_theoretical(self):
        fig_name = "test_msd_mqe_theory.png"
        msd = ManovaSpectrumDistribution(beta=4, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=40,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mre_ratio_ge1(self):
        fig_name = "test_msd_mre_ratio_ge1.png"
        msd = ManovaSpectrumDistribution(beta=1, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=200,
            bins=100,
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_msd_mre_theoretical_ratio_ge1(self):
        fig_name = "test_msd_wre_theory_ratio_ge1.png"
        msd = ManovaSpectrumDistribution(beta=1, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=1000,
            bins=100,
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mre_interval(self):
        fig_name = "test_msd_mre_interval.png"
        msd = ManovaSpectrumDistribution(beta=1, a=3, b=3)
        msd.plot_empirical_pdf(
            m_size=200,
            bins=100,
            interval=(0,10),
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mre_a_le_1_b_le_1(self):
        fig_name = "test_msd_plot_mre_a_le_1_b_le_1.png"
        msd = ManovaSpectrumDistribution(beta=1, a=0.9, b=0.9)
        msd.plot_empirical_pdf(
            m_size=200,
            bins=100,
            interval=(0,10),
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mre_a_le_1_b_le_1_density(self):
        fig_name = "test_msd_plot_mre_a_le_1_b_le_1_density.png"
        msd = ManovaSpectrumDistribution(beta=1, a=0.9, b=0.9)
        msd.plot_empirical_pdf(
            m_size=200,
            bins=100,
            interval=(0,10),
            density=True,
            plot_law_pdf=True,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True

    
    def test_msd_plot_mre_emppdf_n1_size(self):
        fig_name = "test_msd_mre_emppdf_n1_size.png"
        mpd = ManovaSpectrumDistribution(beta=1, a=3, b=3)
        mpd.plot_empirical_pdf(
            m_size=200,
            n1_size=500,
            bins=100,
            interval=(0,10),
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mre_emppdf_n2_size(self):
        fig_name = "test_msd_mre_emppdf_n2_size.png"
        mpd = ManovaSpectrumDistribution(beta=1, a=3, b=3)
        mpd.plot_empirical_pdf(
            m_size=200,
            n2_size=500,
            bins=100,
            interval=(0,10),
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_mre_emppdf_n1n2_size(self):
        fig_name = "test_msd_mre_emppdf_n1n2_size.png"
        mpd = ManovaSpectrumDistribution(beta=1, a=3, b=3)
        mpd.plot_empirical_pdf(
            m_size=200,
            n1_size=500,
            n2_size=600,
            bins=100,
            interval=(0,10),
            density=False,
            plot_law_pdf=False,
            savefig_path=TMP_DIR_PATH+"/"+fig_name
        )
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True
    
    def test_msd_plot_size_exception(self):
        with pytest.raises(ValueError):
            msd = ManovaSpectrumDistribution(beta=1, a=3, b=3)
            msd.plot_empirical_pdf(m_size=0)
    
    def test_msd_plot_ensemble_exception(self):
        with pytest.raises(ValueError):
            msd = ManovaSpectrumDistribution(beta=0, a=3, b=3)
            msd.plot_empirical_pdf(m_size=10)



def test_indicator_func():
    '''Testing indicator function
    '''
    assert _indicator(1.0, start=1.0, stop=2.0, inclusive="both") == 1.0
    assert _indicator(1.0, start=1.0, stop=2.0, inclusive="left") == 1.0
    assert _indicator(1.0, start=1.0, stop=2.0, inclusive="right") == 0.0
    assert _indicator(1.0, start=1.0, stop=2.0, inclusive="neither") == 0.0
    assert _indicator(2.0, start=1.0, stop=2.0, inclusive="both") == 1.0
    assert _indicator(2.0, start=1.0, stop=2.0, inclusive="left") == 0.0
    assert _indicator(2.0, start=1.0, stop=2.0, inclusive="right") == 1.0
    assert _indicator(2.0, start=1.0, stop=2.0, inclusive="neither") == 0.0
    assert _indicator(2.0, stop=2.0, inclusive="both") == 1.0
    assert _indicator(2.0, stop=2.0, inclusive="left") == 0.0
    assert _indicator(2.0, stop=2.0, inclusive="right") == 1.0
    assert _indicator(2.0, stop=2.0, inclusive="neither") == 0.0

def test_indicator_func_except():
    '''Testing indicator function raising exception
    '''
    with pytest.raises(ValueError):
        _ = _indicator(2.0)
    
    with pytest.raises(ValueError):
        _ = _indicator(2.0, start=2.0, inclusive="foo")
    


'''Wishart Ensemble Test module

Testing WishartEnsemble module
'''

import pytest
import numpy as np
from scipy import sparse
from numpy.testing import (
    assert_almost_equal,
    assert_array_equal,
)

from skrmt.ensemble import WishartEnsemble



def test_init_exception():
    with pytest.raises(ValueError):
        _ = WishartEnsemble(beta=3, p=100, n=300)

def test_tridiagonal_ratio_exception():
    with pytest.raises(ValueError):
        _ = WishartEnsemble(beta=1, p=300, n=100, use_tridiagonal=True)

def test_tridiagonal_sigma_exception():
    with pytest.raises(ValueError):
        _ = WishartEnsemble(beta=1, p=100, n=300, sigma=2.0, use_tridiagonal=True)


##########################################
### Wishart Real Ensemble = WRE

def test_wishart_real_init():
    '''Testing WRE init
    '''
    p_size = 3
    n_size = 5

    np.random.seed(1)
    wre = WishartEnsemble(beta=1, p=p_size, n=n_size)

    assert wre.matrix.shape == (p_size,p_size)

    assert_almost_equal(wre.matrix, np.array([[5.1919012, -4.96197148, 5.1988152],
                                            [-4.96197148, 9.08485594, -7.11948641],
                                            [5.1988152, -7.11948641, 7.91882311]]),
                        decimal=7)


def test_wre_symmetric():
    '''Testing that WRE matrix is symmetric
    '''
    p_size = 3
    n_size = 5
    wre = WishartEnsemble(beta=1, p=p_size, n=n_size)

    mtx = wre.matrix
    assert (mtx.transpose() == mtx).all()


def test_wre_set_size():
    '''Testing setter to change matrix sizes of WRE
    '''
    p1_size, n1_size = 3, 5
    p2_size, n2_size = 4, 6

    ens = WishartEnsemble(beta=1, p=p1_size, n=n1_size)
    assert ens.p == p1_size
    assert ens.n == n1_size
    assert ens.matrix.shape == (p1_size,p1_size)

    ens.set_size(p=p2_size, n=n2_size, resample_mtx=False)
    assert ens.p == p2_size
    assert ens.n == n2_size
    assert ens.matrix.shape == (p1_size,p1_size)

    ens.set_size(p=p2_size, n=n2_size, resample_mtx=True)
    assert ens.p == p2_size
    assert ens.n == n2_size
    assert ens.matrix.shape == (p2_size,p2_size)


def test_wre_build_tridiagonal():
    '''Testing tridiagonal form of WRE
    '''
    p_size, n_size = 3, 5
    beta = 1

    # sampling WishartReal tridiagonal
    np.random.seed(1)
    wre = WishartEnsemble(beta=1, p=p_size, n=n_size, use_tridiagonal=True)

    # sampling chi-squares and finding tridiagonal matrix in two ways
    np.random.seed(1)
    a_val = n_size*beta/2
    dfs = np.arange(p_size)
    chisqs_diag = np.array([np.sqrt(np.random.chisquare(2*a_val - beta*df)) for df in dfs])
    dfs = np.flip(dfs)
    chisqs_offdiag = np.array([np.sqrt(np.random.chisquare(beta*df)) for df in dfs[:-1]])
    diagonals = [chisqs_offdiag, chisqs_diag]
    mtx = sparse.diags(diagonals, [-1, 0])
    mtx = mtx.toarray()

    diag = np.array([chisqs_diag[0]**2]+[chisqs_diag[i+1]**2 + chisqs_offdiag[i]**2 \
                     for i in range(p_size-1)])
    offdiag = np.multiply(chisqs_offdiag, chisqs_diag[:-1])

    for i in range(p_size):
        assert diag[i] == wre.matrix[i][i]
    for i in range(p_size-1):
        assert offdiag[i] == wre.matrix[i][i+1]
        assert offdiag[i] == wre.matrix[i+1][i]

    assert_almost_equal(wre.matrix, np.dot(mtx, mtx.transpose()), decimal=7)


def test_beta1_joint_eigval_pdf():
    '''Testing joint eigenvalue pdf
    '''
    p_size, n_size = 3, 5
    wre = WishartEnsemble(beta=1, p=p_size, n=n_size)

    wre.matrix = np.zeros((p_size,p_size))
    assert wre.joint_eigval_pdf() == 0.0

    wre.matrix = np.eye(p_size)
    assert wre.joint_eigval_pdf() == 0.0

    wre.matrix = 10*np.eye(p_size)
    assert wre.joint_eigval_pdf() == 0.0


def test_wre_tridiag_hist():
    '''Testing tridiagonal histogram of WRE
    '''
    p_size, n_size = 50, 100
    wre1 = WishartEnsemble(beta=1, p=p_size, n=n_size, use_tridiagonal=False)
    wre2 = WishartEnsemble(beta=1, p=p_size, n=n_size, use_tridiagonal=True)

    wre1.matrix = wre2.matrix

    nbins = 20
    interval = (0,400)
    to_norm = False # without normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = wre1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = wre2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm)

    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)

    to_norm = True # density normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = wre1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm, normalize=True)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = wre2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm, normalize=True)

    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_almost_equal(hist_nottridiag, hist_tridiag, decimal=7)

    to_norm = False # no density normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = wre1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm, normalize=True)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = wre2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm, normalize=True)

    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)


##########################################
### Wishart Complex Ensemble = WCE

def test_wishart_complex_init():
    '''Testing WCE init
    '''
    p_size = 3
    n_size = 8

    np.random.seed(1)
    wce = WishartEnsemble(beta=2, p=p_size, n=n_size)

    assert wce.matrix.shape == (p_size,p_size)

    mtx_sol = [[17.27142618+0.j, 4.56375544+1.14721231j, \
                -5.16181787+0.9188017j],
               [4.56375544-1.14721231j, 15.53877203+0.j, \
                2.44108976+10.47646539j],
               [-5.16181787-0.9188017j, 2.44108976-10.47646539j, \
                13.83214319+0.j]]

    assert_almost_equal(wce.matrix, np.array(mtx_sol), decimal=7)


def test_wce_hermitian():
    '''Testing that WCE matrix is hermitian
    '''
    p_size = 3
    n_size = 5
    wce = WishartEnsemble(beta=2, p=p_size, n=n_size)

    mtx = wce.matrix
    assert_almost_equal(mtx.transpose().conj(), mtx, decimal=7)


def test_wce_set_size():
    '''Testing setter to change matrix sizes of WCE
    '''
    p1_size, n1_size = 5, 10
    p2_size, n2_size = 7, 14

    ens = WishartEnsemble(beta=2, p=p1_size, n=n1_size)
    assert ens.p == p1_size
    assert ens.n == n1_size
    assert ens.matrix.shape == (p1_size,p1_size)

    ens.set_size(p=p2_size, n=n2_size, resample_mtx=False)
    assert ens.p == p2_size
    assert ens.n == n2_size
    assert ens.matrix.shape == (p1_size,p1_size)

    ens.set_size(p=p2_size, n=n2_size, resample_mtx=True)
    assert ens.p == p2_size
    assert ens.n == n2_size
    assert ens.matrix.shape == (p2_size,p2_size)


def test_wce_build_tridiagonal():
    '''Testing tridiagonal form of WCE
    '''
    p_size, n_size = 3, 5
    beta = 2

    # sampling WishartComplex tridiagonal
    np.random.seed(1)
    wce = WishartEnsemble(beta=2, p=p_size, n=n_size, use_tridiagonal=True)

    # sampling chi-squares and finding tridiagonal matrix in two ways
    np.random.seed(1)
    a_val = n_size*beta/ 2
    dfs = np.arange(p_size)
    chisqs_diag = np.array([np.sqrt(np.random.chisquare(2*a_val - beta*df)) for df in dfs])
    dfs = np.flip(dfs)
    chisqs_offdiag = np.array([np.sqrt(np.random.chisquare(beta*df)) for df in dfs[:-1]])
    diagonals = [chisqs_offdiag, chisqs_diag]
    mtx = sparse.diags(diagonals, [-1, 0])
    mtx = mtx.toarray()

    diag = np.array([chisqs_diag[0]**2]+[chisqs_diag[i+1]**2 + chisqs_offdiag[i]**2 \
                     for i in range(p_size-1)])
    offdiag = np.multiply(chisqs_offdiag, chisqs_diag[:-1])

    for i in range(p_size):
        assert diag[i] == wce.matrix[i][i]
    for i in range(p_size-1):
        assert offdiag[i] == wce.matrix[i][i+1]
        assert offdiag[i] == wce.matrix[i+1][i]

    assert_almost_equal(wce.matrix, np.dot(mtx, mtx.transpose()), decimal=7)


def test_wce_tridiag_hist():
    '''Testing tridiagonal histogram of WCE
    '''
    p_size, n_size = 50, 100
    wce1 = WishartEnsemble(beta=2, p=p_size, n=n_size, use_tridiagonal=False)
    wce2 = WishartEnsemble(beta=2, p=p_size, n=n_size, use_tridiagonal=True)

    wce1.matrix = wce2.matrix

    nbins = 20
    interval = (0,400)
    to_norm = False # without normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = wce1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = wce2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm)

    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)

    to_norm = True # normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = wce1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = wce2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm)

    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_almost_equal(hist_nottridiag, hist_tridiag, decimal=7)



##########################################
### Wishart Quaternion Ensemble = WQE

def test_wishart_quatern_init():
    '''Testing WQE init
    '''
    p_size = 2
    n_size = 5

    np.random.seed(1)
    wqe = WishartEnsemble(beta=4, p=p_size, n=n_size)

    assert wqe.matrix.shape == (2*p_size,2*p_size)

    mtx_sol = [[1.97823453e+01+0.00000000e+00j, -6.78596934e+00-5.77477475e+00j,\
                5.04164630e-17+4.44089210e-16j, 4.98065254e+00+2.42862094e+00j],
               [-6.78596934e+00+5.77477475e+00j, 1.77567967e+01+0.00000000e+00j,\
                -4.98065254e+00-2.42862094e+00j, 4.13628735e-17+4.44089210e-16j],
               [5.04164630e-17-4.44089210e-16j, -4.98065254e+00+2.42862094e+00j,\
                1.97823453e+01+0.00000000e+00j, -6.78596934e+00+5.77477475e+00j],
               [4.98065254e+00-2.42862094e+00j, 4.13628735e-17-4.44089210e-16j,\
                -6.78596934e+00-5.77477475e+00j, 1.77567967e+01+0.00000000e+00j]]

    assert_almost_equal(wqe.matrix, np.array(mtx_sol), decimal=7)


def test_wqe_hermitian():
    '''Testing that WQE matrix is hermitian
    '''
    p_size = 3
    n_size = 5
    wqe = WishartEnsemble(beta=4, p=p_size, n=n_size)

    mtx = wqe.matrix
    assert_almost_equal(mtx, mtx.transpose().conj(), decimal=7)


def test_wqe_set_size():
    '''Testing setter to change matrix sizes of WQE
    '''
    p_size1, n_size1 = 2, 3
    p_size2, n_size2 = 4, 5

    ens = WishartEnsemble(beta=4, p=p_size1, n=n_size1)
    assert ens.p == p_size1
    assert ens.n == n_size1
    assert ens.matrix.shape == (2*p_size1,2*p_size1)

    ens.set_size(p=p_size2, n=n_size2, resample_mtx=False)
    assert ens.p == p_size2
    assert ens.n == n_size2
    assert ens.matrix.shape == (2*p_size1,2*p_size1)

    ens.set_size(p=p_size2, n=n_size2, resample_mtx=True)
    assert ens.p == p_size2
    assert ens.n == n_size2
    assert ens.matrix.shape == (2*p_size2,2*p_size2)


def test_wqe_build_tridiagonal():
    '''Testing tridiagonal form of WQE
    '''
    p_size, n_size = 2, 5
    beta = 4

    # sampling WishartQuaternion tridiagonal
    np.random.seed(1)
    wqe = WishartEnsemble(beta=4, p=p_size, n=n_size, use_tridiagonal=True)

    # sampling chi-squares and finding tridiagonal matrix in two ways
    np.random.seed(1)
    a_val = n_size*beta/ 2
    dfs = np.arange(p_size)
    chisqs_diag = np.array([np.sqrt(np.random.chisquare(2*a_val - beta*df)) for df in dfs])
    dfs = np.flip(dfs)
    chisqs_offdiag = np.array([np.sqrt(np.random.chisquare(beta*df)) for df in dfs[:-1]])
    diagonals = [chisqs_offdiag, chisqs_diag]
    mtx = sparse.diags(diagonals, [-1, 0])
    mtx = mtx.toarray()

    diag = np.array([chisqs_diag[0]**2]+[chisqs_diag[i+1]**2 + chisqs_offdiag[i]**2 \
                     for i in range(p_size-1)])
    offdiag = np.multiply(chisqs_offdiag, chisqs_diag[:-1])

    for i in range(p_size):
        assert diag[i] == wqe.matrix[i][i]
    for i in range(p_size-1):
        assert offdiag[i] == wqe.matrix[i][i+1]
        assert offdiag[i] == wqe.matrix[i+1][i]

    assert_almost_equal(wqe.matrix, np.dot(mtx, mtx.transpose()), decimal=7)


def test_wqe_tridiag_hist():
    '''Testing tridiagonal histogram of WQE
    '''
    p_size, n_size = 50, 100
    wqe1 = WishartEnsemble(beta=4, p=p_size, n=n_size, use_tridiagonal=False)
    wqe2 = WishartEnsemble(beta=4, p=p_size, n=n_size, use_tridiagonal=True)

    wqe1.matrix = wqe2.matrix

    nbins = 20
    interval = (0,400)
    to_norm = False # without normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = wqe1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = wqe2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm)

    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_array_equal(hist_nottridiag, hist_tridiag)

    to_norm = True # normalization
    # calculating histogram using standard naive procedure
    hist_nottridiag, bins_nottridiag = wqe1.eigval_hist(bins=nbins, interval=interval,
                                                        density=to_norm)
    # calculating histogram using tridiagonal procedure
    hist_tridiag, bins_tridiag = wqe2.eigval_hist(bins=nbins, interval=interval,
                                                  density=to_norm)

    assert_array_equal(bins_nottridiag, bins_tridiag)
    assert_almost_equal(hist_nottridiag, hist_tridiag, decimal=7)
